import json
import os

from g3core import G3Core
from g3tables import (
    SWDefinitionTable, PLCCompositionIOTable, VisualizationTable
)

from ._g3core_updater import SWSystemDictG3CoreUpdater
from ._hw_connections import get_hardware_connections
from ._iomap_updater import (
    SWSystemDictControlIOMapUpdater, SWSystemDictTestIOMapUpdater
)
from ._logger import logger  # noqa: F401
from .type_hinting import SystemDict


class SystemConfigProcessor:
    def __init__(
        self,
        system_name: str | None = None,
        sw_table: SWDefinitionTable | None = None,
        io_table: PLCCompositionIOTable | None = None,
        visu_table: VisualizationTable | None = None,
        g3core: G3Core | None = None
    ) -> None:
        self.system_name = system_name
        self.sw_table = sw_table
        self.io_table = io_table
        self.visu_table = visu_table
        self.g3core = g3core
        self._system_dict: SystemDict = {
            'Hardware': [], 'Software': {}, 'Visu': {}
            }

    def _process_sw_table(self) -> None:
        if self.sw_table is None:
            return
        if self.g3core is None:
            raise ValueError('G3Core not provided')
        if not self.system_name:
            raise ValueError('System name not provided')
        sw_system_dict = self.sw_table.to_dict(self.system_name)
        SWSystemDictG3CoreUpdater(sw_system_dict, self.g3core).validate()
        SWSystemDictTestIOMapUpdater(sw_system_dict, self.g3core).update()
        self._system_dict['Software'] = sw_system_dict

    def _process_io_table(self) -> None:
        if self.io_table is None:
            return
        if self._system_dict['Software']:
            SWSystemDictControlIOMapUpdater(
                self._system_dict['Software'],
                iomap=self.io_table.get_iomapping(),
                sl8101_name=self.io_table.get_sl8101_name(),
                g3core=self.g3core
            ).update()
        self._system_dict['Hardware'] = get_hardware_connections(self.io_table)

    def _process_visu_table(self) -> None:
        if self.visu_table is None:
            return
        table = self.visu_table
        visu = table.project_types
        if '' in visu:
            del visu['']
        for data in [
            table.gate_tables,
            table.heating_tables,
            table.requestor_tables
        ]:
            for type_, subtype_data in data.items():  # type: ignore
                type_data = visu.setdefault(
                    type_,
                    {'type': type_, 'restricted_to': None, 'vars': {}}
                    )
                type_data['type_variants'] = subtype_data
        self._system_dict['Visu'] = visu

    def to_dict(self) -> SystemDict:
        self._process_sw_table()
        self._process_io_table()
        self._process_visu_table()
        return self._system_dict

    def to_file(
        self,
        dirpath: str = '.',
        filename: str = 'SystemConfig.json',
        update_if_exists: bool = False
    ) -> None:
        dirpath = dirpath if dirpath != '.' else os.getcwd()
        system_dict = self.to_dict()
        filepath = os.path.join(dirpath, filename)
        if not update_if_exists or not os.path.exists(filepath):
            with open(filepath, 'w', encoding='utf-8') as file:
                json.dump(system_dict, file, indent=4)
        elif not os.path.isfile(filepath):
            raise TypeError(
                f'Path "{filepath}" exists, but does not lead to a file.'
                )
        else:
            with open(filepath, 'r', encoding='utf-8') as file:
                system_dict_old = json.load(file)
            if not isinstance(system_dict_old, dict):
                raise TypeError(
                    f'System configuration file "{filepath}" has unexpected '
                    f'structure.'
                    )
            system_dict_old.update(system_dict)
            with open(filepath, 'w', encoding='utf-8') as file:
                json.dump(system_dict_old, file, indent=4)



import gspread  # type: ignore
import json
import os
import logging
import pandas
import shutil
import typing
import pandas as pd

from platformdirs import user_config_dir


logger = logging.getLogger('g3tables.utils.gdrive')


class CredsFileManager:
    def __init__(self):
        config_dir = user_config_dir('g3tables', ensure_exists=True)
        self._file_path = os.path.join(config_dir, "creds.json")

    def validate_creds_file(self, creds_file_path: str) -> None:
        if os.path.exists(creds_file_path) is False:
            raise FileNotFoundError(f'File not found: "{creds_file_path}"')
        with open(creds_file_path, 'r') as file:
            try:
                json.load(file)
            except Exception as exp:
                raise ValueError(f'File could not be read: {exp}') from exp

    def update_creds_file(self, creds_file_path: str) -> None:
        self.validate_creds_file(creds_file_path)
        if creds_file_path != self._file_path:
            shutil.copyfile(creds_file_path, self._file_path)

    def update_creds_file_from_input(self) -> None:
        try:
            file_path = input(
                'Please provide the path to the Google Drive '
                'service account credentials file: '
                )
            self.update_creds_file(file_path)
            print('Credentials file has been updated.')
        except KeyboardInterrupt:
            print('\nCredentials file update has been aborted.')
            exit()
        except Exception as exp:
            print(f'Failed to update credentials file ({exp}).\n')
            self.update_creds_file_from_input()

    def get_creds_file(self, ensure_is_valid: bool = True) -> str:
        if ensure_is_valid is False:
            return self._file_path
        try:
            self.validate_creds_file(self._file_path)
            return self._file_path
        except Exception:
            self.update_creds_file_from_input()
            return self._file_path


def get_gspread_client(creds_file_path: str | None = None) -> gspread.Client:
    logger.info('Loading Google Drive service account credentials file')
    creds_file_manager = CredsFileManager()
    if creds_file_path:
        creds_file_manager.update_creds_file(creds_file_path)
    creds_file = creds_file_manager.get_creds_file()
    logger.info('Connecting to Google Drive repository')
    return gspread.service_account(filename=creds_file)


class TableMetadataDict(typing.TypedDict):
    id: str
    name: str
    createdTime: str
    modifiedTime: str


def list_tables_from_gdrive_with_meta(
    project: str | typing.Any | None
) -> list[TableMetadataDict]:
    client = get_gspread_client()
    logger.info('Collecting Google Drive G3 project tables')
    tables: list[TableMetadataDict] = [
        table for table in client.list_spreadsheet_files()
        ]
    if project is not None:
        logger.info('Filtering "%s" Google Drive G3 project tables', project)
        tables = [
            table for table in tables
            if str(project).casefold() in str(table['name']).casefold()
            ]
    return tables


def list_tables_from_gdrive(project: str | typing.Any | None) -> list[str]:
    tables = list_tables_from_gdrive_with_meta(project)
    return [table['name'] for table in tables]


def get_table_data_from_gdrive(
    gdrive_table_name: str,
    header_row: int = 1,
    include_sheets: typing.Optional[typing.Iterable[str]] = None,
    exclude_sheets: typing.Optional[typing.Iterable[str]] = None
) -> dict[str, pandas.DataFrame]:
    """both excluded and included == excluded"""
    client = get_gspread_client()
    try:
        logger.info('Loading spreadsheet "%s"', gdrive_table_name)
        table = client.open(gdrive_table_name)
    except gspread.exceptions.SpreadsheetNotFound:
        logger.error(
            'Unable to access spreadsheet "%s". Make sure '
            'that the table name is spelled correctly and '
            'that access to the table has been granted to '
            '"el-script@g3-spreadsheet-sync.iam.gserviceaccount.com".',
            gdrive_table_name
            )
        return {}
    sheets = {}
    for sheet in table.worksheets(exclude_hidden=True):
        if (
            (exclude_sheets and sheet.title in exclude_sheets) or
            (include_sheets and sheet.title not in include_sheets)
        ):
            logger.debug('Ignoring sheet "%s"', sheet.title)
            continue
        try:
            logger.info('Loading sheet "%s"', sheet.title)
            df = pd.DataFrame(
                sheet.get_all_records(
                    head=header_row, numericise_ignore=['all']
                )
            )

            df = df[~df.iloc[:, 4].str.contains("The script does not work with this line", na=False)]
            sheets[sheet.title] = df.reset_index(drop=True)  # Reset indexu

        except Exception as exp:
            if logger.isEnabledFor(logging.WARNING):  # optimize err str format
                err = f'Failed to load sheet "{sheet.title}"'
                if (exp_str := str(exp)):
                    err = f'{err} ({exp_str})'
                logger.warning(err)
    return sheets



from ._gdrive import (
    CredsFileManager,
    get_gspread_client,
    list_tables_from_gdrive,
    list_tables_from_gdrive_with_meta,
    get_table_data_from_gdrive
    )


__all__ = [
    'CredsFileManager',
    'get_gspread_client',
    'list_tables_from_gdrive',
    'list_tables_from_gdrive_with_meta',
    'get_table_data_from_gdrive'
    ]



import argparse
import os

from ._gdrive import CredsFileManager


def main():
    manager = CredsFileManager()
    creds_dir = os.path.dirname(manager.get_creds_file(ensure_is_valid=False))
    parser = argparse.ArgumentParser(
        description=(
            f'Update Google Drive service account credentials file.\n'
            f'If the contents of the file are a valid JSON object, '
            f'they are copied to the directory of the g3tables library: '
            f'"{creds_dir}".'
            )
        )
    parser.add_argument(
        "creds_file_path",
        nargs="?",
        type=str,
        help=(
            "Path to the new Google Drive service account credentials file. "
            "If not provided, the user will be prompted to input "
            "the file path via the console."
            )
        )
    args = parser.parse_args()
    path = args.creds_file_path
    if path:
        try:
            manager.update_creds_file(path)
        except Exception as exp:
            print(f'Failed to update the credentials file ({exp}).\n')
            manager.update_creds_file_from_input()
    else:
        manager.update_creds_file_from_input()


if __name__ == "__main__":
    main()



import itertools

from g3core import G3Core, STVariable

from .type_hinting import (
    SWDeviceDict,
    SWModuleDict,
    SWSystemDict
)
from ._sw_system_dict_wrapper import SWSystemDictWrapper
from ._logger import logger


class SWSystemDictControlIOMapUpdater:
    EXCLUDED_FOR_SAFETY_IOMAP = []

    def __init__(
        self,
        system: SWSystemDict,
        iomap: dict[str, dict[str, str]] | None,
        sl8101_name: str | None,
        g3core: G3Core | None
    ) -> None:
        self.wrapper = SWSystemDictWrapper(system)
        self.iomap = iomap
        self.sl8101_name = sl8101_name
        self.g3core = g3core
        self._counters = self._create_iomap_counters()

    @staticmethod
    def _create_iomap_counters() -> dict:
        return {
            "inFromSafety": {
                "UDINT": itertools.count(1),
                "UINT": itertools.count(1)
            },
            "outToSafety": {
                "UDINT": itertools.count(1001),
                "UINT": itertools.count(1001)
            },
            "inAuthKey": {
                "INT": itertools.count(1)
            },
            "inAuthBusy": {
                "BOOL": itertools.count(1)
            },
            "inAuthResult": {
                "BOOL": itertools.count(2)
            },
            "outAuthKey": {
                "INT": itertools.count(1001)
            },
            "outAuthCmd": {
                "INT": itertools.count(1002)
            }
        }

    def _update_iomap_device(
        self, device_data: SWDeviceDict, iomap_key: str
    ) -> None:
        if not self.iomap:
            return
        iomap = self.iomap.get(iomap_key)
        if iomap is None:
            return
        iomap_dict = device_data['control'].setdefault('iomap', {})
        for key, value in iomap.items():
            iomap_dict[key] = value

    def _generate_safety_hardware_mapping(self, var: STVariable) -> str:
        assert self.sl8101_name is not None
        match var.name, var.dtype:
            case 'inFromSafety', 'UINT':
                prefix = '%IW'
            case 'inFromSafety', 'UDINT':
                prefix = '%ID'
            case 'inAuthKey', 'INT':
                prefix = '%IW'
            case 'inAuthBusy', 'BOOL':
                prefix = '%IX'
            case 'inAuthResult', 'BOOL':
                prefix = '%IX'
            case 'outAuthKey', 'INT':
                prefix = '%QW'
            case 'outAuthCmd', 'INT':
                prefix = '%QW'
            case 'outToSafety', 'UINT':
                prefix = '%QW'
            case 'outToSafety', 'UDINT':
                prefix = '%QD'
            case _:
                logger.warning(
                    'Unable to generate safety IO mapping for variable "%s" '
                    'of type "%s"', var.name, var.dtype
                    )
                return ''
        counter = self._counters[var.name][var.dtype]
        if (var.name == "inAuthBusy") or (var.name == "inAuthResult") :
            return (
                f'{prefix}."{self.sl8101_name}".{var.dtype}'
                f'{next(counter):05}'
                )
        else:
            return (
                f'{prefix}."{self.sl8101_name}".{var.dtype}'
                f'{next(counter):04}'
                )

    def _update_iomap_device_safety(
        self, device_type: str, device_data: SWDeviceDict
    ) -> None:
        device_name = self.wrapper.get_device_name(device_data)
        # get the control fb name
        fb_name = self.wrapper.get_device_control_fb(device_data)
        if not fb_name:
            if not self.sl8101_name:
                return
            logger.warning(
                '%s "%s" function block name was not found in the config data.'
                ' safety IO mapping to harware module "%s" cannot be added.',
                device_type, device_name, self.sl8101_name
                )
            return
        # check if device is a safety device (from fb name)
        if 'safety' not in fb_name.lower() or 'non' in fb_name.lower():
            logger.debug(
                '%s "%s" is not a safety device. safety IO mapping to harware '
                'module "%s" is not needed.',
                device_type, device_name, self.sl8101_name
                )
            return
        # check if SL8101 module and G3 Core library data is provided
        if self.sl8101_name is None:
            logger.warning(
                'Unable to generate safety IO mapping for %s "%s" '
                '(X20cSL8101 harware module name was not found).',
                device_type, device_name
                )
            return
        logger.info(
            'Generating safety IO mapping to harware module "%s" for '
            '%s "%s".', self.sl8101_name, device_type, device_name
            )
        if self.g3core is None:
            logger.warning(
                'Unable to generate safety IO mapping for %s "%s" '
                '(G3 Core library files were not found).',
                device_type, device_name
                )   # G3 Core check is inside this func because it may not be
            return  # needed if the project is not safety
        # find the fb data
        try:
            _, control_fb = self.g3core.find_function_block(fb_name)
        except ValueError as err:
            logger.warning(
                'Unable to generate safety IO mapping for %s "%s" (%s).',
                device_type, device_name, str(err)
                )
            return
        # generate and fill in the safety module io mapping
        iomap_dict = device_data['control'].setdefault('iomap', {})
        for var in control_fb.var_input:
            if var.name == "inFromSafety":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "inFromSafety" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['inFromSafety'] = hardware_signal
                    break
        for var in control_fb.var_out:
            if var.name == "outToSafety":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "outToSafety" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['outToSafety'] = hardware_signal
                    break
        for var in control_fb.var_input:
            if var.name == "inAuthKey":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "inAuthKey" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['inAuthKey'] = hardware_signal
                    break 
        for var in control_fb.var_input:
            if var.name == "inAuthBusy":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                print(hardware_signal)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "inAuthBusy" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['inAuthBusy'] = hardware_signal
                    break        
        for var in control_fb.var_input:
            if var.name == "inAuthResult":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                print(hardware_signal)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "inAuthResult" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['inAuthResult'] = hardware_signal
                    break    
        for var in control_fb.var_out:
            if var.name == "outAuthKey":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "outAuthKey" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['outAuthKey'] = hardware_signal
                    break    
        for var in control_fb.var_out:
            if var.name == "outAuthCmd":
                hardware_signal = self._generate_safety_hardware_mapping(var)
                if hardware_signal:
                    logger.info(
                        'Mapping %s "%s" "outAuthCmd" connector to "%s".',
                        device_type, device_name, hardware_signal
                        )
                    iomap_dict['outAuthCmd'] = hardware_signal
                    break    

    def _update_iomap_module(
        self,
        module_name: str,
        module_data: SWModuleDict,
        ignore: list[str] | None = None
    ) -> None:
        for device_name, device_data in module_data.items():
            if ignore is None or module_name not in ignore:
                iomap_key = f'{module_name}/{device_name}'
                self._update_iomap_device(device_data, iomap_key)
                self._update_iomap_device_safety(module_name, device_data)
            children = device_data.get('children')
            if children:
                for ch_module_name, ch_module_data in children.items():
                    self._update_iomap_module(
                        f'{module_name}/{ch_module_name}',
                        ch_module_data,
                        ignore=ignore
                        )

    def _update_iomap_module_cabinet(
        self,
        module_data: SWModuleDict,
        ignore: list[str] | None = None
    ) -> None:
        for cabinet_name, cabinet_data in module_data.items():
            iomap_key = f'Cabinet/{cabinet_name}'
            if ignore is None or 'Cabinet' not in ignore:
                self._update_iomap_device(cabinet_data, iomap_key)
            children = cabinet_data.get('children')
            assert children is not None
            for ch_module_name, ch_module_data in children.items():
                if ignore and ch_module_name in ignore:
                    continue
                for ch_device_name, ch_device_data in ch_module_data.items():
                    ch_iomap_key = (
                        f'{iomap_key}/{ch_module_name}/{ch_device_name}'
                        )
                    self._update_iomap_device(ch_device_data, ch_iomap_key)
                    self._update_iomap_device_safety(
                        ch_module_name, ch_device_data
                        )

    def update(self) -> None:
        for zone_name, zone_data in self.wrapper.system_dict.items():
            if zone_name == 'Common':  # nothing to update in the 'Common' data
                for module_name, module_data in zone_data.items():
                    if module_name == 'SystemSafety':
                        self._update_iomap_module(
                            module_name,
                            module_data,
                            ignore=self.EXCLUDED_FOR_SAFETY_IOMAP
                            )
                continue
            for module_name, module_data in zone_data.items():
                if module_name == 'Cabinet':
                    self._update_iomap_module_cabinet(module_data)
                else:
                    self._update_iomap_module(
                        module_name,
                        module_data,
                        ignore=self.EXCLUDED_FOR_SAFETY_IOMAP
                        )


class SWSystemDictTestIOMapUpdater:
    def __init__(self, system: SWSystemDict, g3core: G3Core) -> None:
        self.wrapper = SWSystemDictWrapper(system)
        self.g3core = g3core

    def update(self) -> None:
        pass



import enum
import logging
import numpy as np
import pandas as pd
import re
import typing

from openpyxl.utils import get_column_letter

from ..utils.gdrive import get_table_data_from_gdrive


logger = logging.getLogger('g3tables.plc_composition_io_table')


class PLCCompositionIOTableColumn(enum.StrEnum):
    """
    Enumeration of column names in a PLC Composition IO table. This enum
    facilitates easier and more readable access to specific columns
    within the table data.

    Each enum member represents a column name. The docstring for each member
    includes a brief description and indicates its position in both
    the original and the expanded table. The position is denoted in the format
    (original_index_numeric/original_index_alphabet,
    expanded_index_numeric/expanded_index_alphabet).
    """
    CABINET = "Cabinet"
    """
    The name of the cabinet, typically the same as the sheet name.
    This column exists only in the expanded table.
    Position in tables: Original (-), Expanded (1/A).
    """
    COUNT = "Count"
    """
    The count of PLC units in the cabinet. Originally, this is the first
    unnamed column. Position in tables: Original (1/A), Expanded (2/B).
    """
    PLC_UNIT = "PLC Unit / from left to right /"
    """
    The name of the PLC hardware unit.
    Position in tables: Original (2/B), Expanded (3/C).
    """
    TERMINAL = "Terminal at the module"
    """
    The terminal location at the PLC module.
    Position in tables: Original (3/C), Expanded (4/D).
    """
    LED_INDICATOR = "LED indicator at the module"
    """
    The LED indicator located on the PLC module.
    Position in tables: Original (4/D), Expanded (5/E).
    """
    SLOT_ST = "Slot ST"
    """
    The Slot ST identifier.
    Position in tables: Original (5/E), Expanded (6/F).
    """
    SIGNAL_NAME = "Signal name at PLC module"
    """
    The name of the signal at the PLC hardware module, such as
    'DigitalInput01'. Position in tables: Original (6/F), Expanded (7/G).
    """
    POLARITY = "Polarity of OUT /Pulse x or +24V for IN"
    """
    Describes the polarity for output signals or the voltage for input signals.
    Position in tables: Original (7/G), Expanded (8/H).
    """
    SW_DEVICE_TYPE = "SWDeviceType"
    """
    The type of the software device associated with a PLC hardware module
    signal, such as 'Cabinet/Fuse'. This column exists only in the expanded
    table. Position in tables: Original (-), Expanded (9/I).
    """
    SW_DEVICE_NAME = "SWDeviceName"
    """
    The name of the software device associated with a PLC hardware module
    signal, such as 'FU_01'. This column exists only in the expanded table.
    Position in tables: Original (-), Expanded (10/J).
    """
    SW_SIGNAL = "SWSignal"
    """
    The software signal mapped to a PLC hardware module signal, as defined in
    G3Core function block, such as 'inFuseOk'. This column exists only in
    the expanded table. Position in tables: Original (-), Expanded (11/K).
    """
    SIGNAL = "Signal"
    """
    The software signal mapped to a PLC hardware module signal, as shown on
    the electrical connections schema, such as 'FU1_ST'.
    Position in tables: Original (8/H), Expanded (12/L).
    """
    REMARK = "Remark"
    """
    Any comments or notes related to the signal mapping.
    Position in tables: Original (9/I), Expanded (13/M).
    """
    PLC_UNIT_SUFFIX = "PLCUnitSuffix"
    """
    A unique letter identifier for a plc module. Derived from the column
    "Terminal at the module" (the latter cannot be used directly due to
    the cases when one module has two assigned letters, e.g. X20cSI9100).
    Position in tables: Original (-), Expanded (14/N).
    * TODO: move this column to "C".
    """


PLCCompositionIOTableColumnIndex: dict[str, int] = {
    memb: i + 1 for i, memb in enumerate(PLCCompositionIOTableColumn)
    }


class PLCCompositionIOTable:
    """
    `PLCCompositionIOTable` is a representation of the G3 Project PLC
    Composition and IO table. It allows access to the data of the hardware
    cabinet sheets in the form of a `pandas.DataFrame`.

    Note that the "razitko", "TSC-NxG_T11", "TSC-NxG_T22", "TSC-NxG_T22_20",
    "template units", "units", "objedn", "List1", and "power" sheets are not
    represented within the dictionary.
    """
    NAME_PATTERN = 'plc_composition'
    """
    The pattern that the name of the SW Definition table must match to be
    automatically recognized in a Google Drive folder.
    """
    HEADER_ROW = 11
    """
    The row index of the header row in the table.
    """
    EXCLUDE_SHEETS = [
        'razitko',
        'TSC-NxG_T11',
        'TSC-NxG_T22',
        'TSC-NxG_T22_20',
        'template units',
        'units',
        'objedn',
        'List1',
        'power'
        ]
    """
    Sheets that are not represented in the data dictionary of
    a `PLCCompositionIOTable` instance.
    """

    def __init__(self, data: pd.DataFrame) -> None:
        """
        Standard initializer of a `PLCCompositionIOTable` instance. Although it
        can be used directly, alternative class constructors `from_local` and
        `from_gdrive` are adviced instead.

        Args:
            data (dict[str, pd.DataFrame]): a dict, where keys are\
            sheets names and values are the `pandas.DataFrame` objects\
            representing the sheet data.
        """
        self.data = data

    @classmethod
    def is_name_valid(cls, name: str) -> bool:
        """
        Check if the provided name matches the PLC Composition and IO table
        name pattern.

        Args:
            name (str): The name to validate.

        Returns:
            bool: True if the name is valid, False otherwise.
        """
        name_formatted = name.strip().replace(' ', '_').casefold()
        return cls.NAME_PATTERN in name_formatted

    @classmethod
    def from_local(
        cls, path: str, sheet_name: str | typing.Iterable[str]
    ) -> typing.Self:
        """
        Create a `PLCCompositionIOTable` instance from a local Excel file.

        Args:
            local_path (str): The local file path to the Excel file containing\
                the PLC Composition and IO table data.

        Returns:
            Self: An instance of `PLCCompositionIOTable` initialized\
                with data from the specified local Excel file.
        """
        if isinstance(sheet_name, str):
            sheet_name = [sheet_name]
        else:
            sheet_name = [name for name in sheet_name]
        logger.info(
            'Reading PLC Composition IO table cabitet "%s" data from local '
            'path: "%s"', ", ".join(sheet_name), path
            )
        sheets: dict[str, pd.DataFrame] = pd.read_excel(
            path,
            sheet_name,  # type: ignore
            skiprows=[idx for idx in range(cls.HEADER_ROW - 1)],
            )
        # check if additional columns are inserted
        to_rename = {'Unnamed: 0': PLCCompositionIOTableColumn.COUNT}
        for cabinet_name, sheet in sheets.items():
            PLCCompositionIOTableFormatter.check_inserted_cols(
                cabinet_name, sheet
                )
            sheet.rename(columns=to_rename, inplace=True)
        # format and concat sheet dict to one dataframe
        formatter = PLCCompositionIOTableFormatter.format_and_concat_sheets
        sheets_concat = formatter(sheets)
        return cls(sheets_concat)

    @classmethod
    def from_gdrive(
        cls, gdrive_table_name: str, sheet_name: str | typing.Iterable[str]
    ) -> typing.Self:
        """
        Create a `PLCCompositionIOTable` instance from a Google Drive file.

        Args:
            gdrive_table_name (str): The name of the table in Google Drive\
                containing the PLC Composition and IO table data.

        Returns:
            Self: An instance of `PLCCompositionIOTable` initialized\
                with data from the specified Google Drive file.
        """
        if isinstance(sheet_name, str):
            sheet_name = [sheet_name]
        logger.info(
            'Reading PLC Composition IO table cabitet "%s" data from Google '
            'Drive: "%s"', ", ".join(sheet_name), gdrive_table_name
            )
        try:
            sheets: dict[str, pd.DataFrame] = get_table_data_from_gdrive(
                gdrive_table_name,
                header_row=cls.HEADER_ROW,
                exclude_sheets=cls.EXCLUDE_SHEETS
                )
        except ValueError as err:
            logger.error('%s', str(err))
            exit()
        # check if additional columns are inserted
        to_rename = {'': PLCCompositionIOTableColumn.COUNT}
        for cabinet_name, sheet in sheets.items():
            PLCCompositionIOTableFormatter.check_inserted_cols(
                cabinet_name, sheet
                )
            sheet.rename(columns=to_rename, inplace=True)
        # format and concat sheet dict to one dataframe
        formatter = PLCCompositionIOTableFormatter.format_and_concat_sheets
        sheets_concat = formatter(sheets, include_only=sheet_name)
        return cls(sheets_concat)

    @classmethod
    def load(
        cls, path: str, sheet_name: str | typing.Iterable[str]
    ) -> typing.Self:
        """
        Load a SW Definition table from a local file or Google Drive.

        Args:
            path (str): Path to the SW Definition table file. If the path ends\
                with '.xlsx', the table is loaded from a local file.\
                Otherwise, the table is loaded from Google Drive.

        Returns:
            typing.Self: An instance of `SWDefinitionTable` initialized with\
                data from the specified file.
        """
        if path.endswith('.xlsx'):
            return cls.from_local(path, sheet_name)
        return cls.from_gdrive(path, sheet_name)

    @staticmethod
    def get_column_index(column: str | PLCCompositionIOTableColumn) -> int:
        """
        Get the index of a column in the PLC Composition and IO table.

        Args:
            column (str | PLCCompositionIOTableColumn): The column name or\
                enumeration member.

        Returns:
            int: The index of the column in the table.
        """
        column = PLCCompositionIOTableColumn(column)
        return PLCCompositionIOTableColumnIndex[column]

    @staticmethod
    def get_column_letter(column: str | PLCCompositionIOTableColumn) -> str:
        """
        Get the column letter of a column in the PLC Composition and IO table.

        Args:
            column (str | PLCCompositionIOTableColumn): The column name or\
                enumeration member.

        Returns:
            str: The column letter of the column in the table.
        """
        column = PLCCompositionIOTableColumn(column)
        return get_column_letter(PLCCompositionIOTableColumnIndex[column])

    @staticmethod
    def _get_plc_unit_type(row) -> str | None:
        unit_type = row[PLCCompositionIOTableColumn.PLC_UNIT]
        if PLCCompositionIOTableFormatter.is_plc_unit_type(unit_type):
            return unit_type
        return None

    @staticmethod
    def _get_plc_unit_name(
        unit_type: str, cabinet: str, name_suffix: str
    ) -> str:
        return f'{unit_type}_{cabinet}_{name_suffix}'

    @staticmethod
    def _extract_prefix(s: str) -> str:
        match = re.match(r'^[A-Za-z]+', s)
        return match.group(0) if match else ''

    def _is_terminal_equivalent(self, terminal1: str, terminal2: str) -> bool:
        prefix1 = self._extract_prefix(terminal1)
        prefix2 = self._extract_prefix(terminal2)
        return prefix1 == prefix2

    def iter_plc_units(self) -> typing.Iterator[tuple[str, str, str]]:
        """
        Iterate over the PLC units in the table.

        Yields:
            typing.Iterator[tuple[str, str, str]]: An iterator over the PLC\
                units in the table. Each iteration yields a tuple containing\
                the PLC unit type, the cabinet name, and the terminal suffix.
        """
        suffix_prev = ''
        for _, row in self.data.iterrows():
            plc_unit_type = self._get_plc_unit_type(row)
            if not plc_unit_type:
                continue
            suffix = row[PLCCompositionIOTableColumn.PLC_UNIT_SUFFIX]
            cabinet = row[PLCCompositionIOTableColumn.CABINET]
            if suffix == suffix_prev:
                continue
            else:
                suffix_prev = suffix
            yield plc_unit_type, cabinet, suffix

    def get_sl8101_name(self) -> str | None:
        """
        Get the name of the "X20cSL8101" PLC unit.

        Returns:
            str | None: The name of the "X20cSL8101" PLC unit if found,\
                otherwise None.
        """
        logger.info('Searching for the "X20cSL8101" PLC unit')
        for unit_type, cabinet, terminal in self.iter_plc_units():
            if 'sl8101' in unit_type.lower():
                sl8101 = self._get_plc_unit_name(unit_type, cabinet, terminal)
                logger.debug('Found the "X20cSL8101" PLC unit "%s"', sl8101)
                return sl8101
        logger.warning('Could not find the "X20cSL8101" PLC unit')
        return None

    def get_iomapping(self) -> dict[str, dict[str, str]]:
        """
        Get the IO mapping data from the PLC Composition and IO table.

        Returns:
            dict[str, dict[str, str]]: A dictionary containing the IO mapping\
                data. The keys are the software device type and name, and\
                the values are dictionaries containing the software signal\
                mapped to the hardware signal.
        """

        def get_i_prefix(signal_name: str) -> str:
            logger.debug(
                'Matching hardware signal "%s" IO mapping prefix', signal_name
                )
            name = signal_name.lower()
            if "digitalinput" in name:
                return "%IX"
            if "digitaloutput" in name:
                return "%QX"
            if "analoginput" in name:
                return "%IW"
            if "analogoutput" in name:
                return "%QW"
            if "temperature" in name:
                return "%IW"
            logger.warning('Unexpected hardware signal: "%s"', signal_name)
            return "%UNDEFINED_PREFIX"

        logger.info('Collecting IO mapping data')
        mapping: dict[str, dict[str, str]] = {}
        for _, row in self.data.iterrows():
            cabinet = row[PLCCompositionIOTableColumn.CABINET]
            sw_device_type = row[PLCCompositionIOTableColumn.SW_DEVICE_TYPE]
            if 'Cabinet/' in str(sw_device_type):
                logger.debug(
                    'Evaluated device "%s" as a Cabinet device or a Cabinet '
                    'child device', sw_device_type
                    )
                child_device_type = "/".join(sw_device_type.split('/')[1:])
                sw_device_type = f'Cabinet/{cabinet}/{child_device_type}'
            sw_device_name = row[PLCCompositionIOTableColumn.SW_DEVICE_NAME]
            sw_signal = row[PLCCompositionIOTableColumn.SW_SIGNAL]
            if not sw_device_type or not sw_device_name or not sw_signal:
                continue
            logger.debug(
                'Formatting IO mapping data for device "%s"', sw_device_type
                )
            name_suffix = self._extract_prefix(
                row[PLCCompositionIOTableColumn.TERMINAL]
                )
            hw_unit_type = row[PLCCompositionIOTableColumn.PLC_UNIT]
            hw_module_name = self._get_plc_unit_name(
                hw_unit_type, cabinet, name_suffix
                )
            hw_signal_name = row[PLCCompositionIOTableColumn.SIGNAL_NAME]
            prefix = get_i_prefix(hw_signal_name)
            hw_signal = f'{prefix}."{hw_module_name}".{hw_signal_name}'
            mapping_key = f'{sw_device_type}/{sw_device_name}'
            logger.debug(
                'Found IO mapping data for device "%s": "%s" -> "%s"',
                mapping_key, sw_signal, hw_signal
                )
            device_mapping = mapping.setdefault(mapping_key, dict())
            device_mapping[sw_signal] = hw_signal
        return mapping


class PLCCompositionIOTableFormatter:

    @staticmethod
    def check_inserted_cols(sheet_name: str, sheet: pd.DataFrame):
        cols = [
            PLCCompositionIOTableColumn.SW_DEVICE_TYPE,
            PLCCompositionIOTableColumn.SW_DEVICE_NAME,
            PLCCompositionIOTableColumn.SW_SIGNAL
        ]
        for col in cols:
            logger.info(
                'Checking if column "%s" is present in sheet "%s"',
                col, sheet_name
                )
            if col.value not in sheet.columns:
                raise AttributeError(
                    f'Column "{col.value}" was not found in '
                    f'table sheet "{sheet_name}".'
                    )

    @staticmethod
    def is_plc_unit_type(value) -> bool:
        if not value:
            return False
        value = str(value).strip()
        res = ('X20' in value) and value.isalnum()
        return res

    @staticmethod
    def format_to_int_str(val: typing.Any) -> str | typing.Any:
        # check if the value is numeric and not NaN
        if pd.notna(val) and np.issubdtype(type(val), np.number):
            return str(int(val))
        return val  # return the value unchanged if not numeric or is NaN

    @classmethod
    def add_plc_unit_suffix_col(cls, sheet: pd.DataFrame) -> pd.DataFrame:
        col_name_unit = PLCCompositionIOTableColumn.PLC_UNIT
        col_name_suff = PLCCompositionIOTableColumn.PLC_UNIT_SUFFIX
        col_name_term = PLCCompositionIOTableColumn.TERMINAL
        # get mask of the col values where row does not contain plc unit name
        is_plc_unit_mask = sheet[col_name_unit].apply(cls.is_plc_unit_type)
        # format the suffix column
        sheet[col_name_suff] = None
        term_letters = sheet.loc[is_plc_unit_mask, col_name_term]
        sheet.loc[is_plc_unit_mask, col_name_suff] = term_letters
        sheet[col_name_suff] = sheet[col_name_suff].ffill()
        return sheet

    @classmethod
    def format_plc_unit_col(cls, sheet: pd.DataFrame) -> pd.DataFrame:
        col_name = PLCCompositionIOTableColumn.PLC_UNIT
        # get mask of the col values where row does not contain plc unit name
        is_plc_unit_mask = sheet[col_name].apply(cls.is_plc_unit_type)
        # set non-module plc unit cells to None
        sheet.loc[~is_plc_unit_mask, col_name] = None
        # forward fill the module names
        sheet[col_name] = sheet[col_name].ffill()
        return sheet

    @classmethod
    def format_sheet(cls, sheet: pd.DataFrame) -> pd.DataFrame:
        logger.info('Dropping all empty rows')
        sheet = sheet.replace('', np.nan)
        sheet = sheet.dropna(axis=0, how='all')
        logger.debug(
            'Updating "%s" column values',
            PLCCompositionIOTableColumn.PLC_UNIT
            )
        sheet = cls.add_plc_unit_suffix_col(sheet)
        sheet = cls.format_plc_unit_col(sheet)
        logger.debug('Converting all cell values string representation')
        sheet = sheet.applymap(cls.format_to_int_str)
        sheet = sheet.where(pd.notna(sheet), '')
        sheet = sheet.fillna('')
        sheet = sheet.applymap(lambda cell: str(cell).strip())
        return sheet

    @classmethod
    def format_and_concat_sheets(
        cls,
        sheets: dict[str, pd.DataFrame],
        ignore: typing.Iterable[str] | None = None,
        include_only: typing.Iterable[str] | None = None
    ) -> pd.DataFrame:
        if ignore is None:
            ignore = []
        if include_only is None:
            include_only = [name for name in sheets.keys()]
        sheets_concat = pd.DataFrame()
        logger.info('Formatting and concatenating cabinet sheets')
        for cabinet_name, sheet in sheets.items():
            if cabinet_name in ignore or cabinet_name not in include_only:
                logger.debug('Ignoring sheet "%s"', cabinet_name)
                continue
            logger.info('Formatting sheet "%s"', cabinet_name)
            sheet = cls.format_sheet(sheet)
            # insert 'Cabinet' column
            logger.info(
                'Insering column "%s" into sheet "%s"',
                PLCCompositionIOTableColumn.CABINET, cabinet_name
                )
            sheet.insert(1, 'Cabinet', cabinet_name)
            # concatenate sheets
            logger.info('Concatenting sheet "%s"', cabinet_name)
            to_concat = [sheets_concat, sheet]
            sheets_concat = pd.concat(to_concat, ignore_index=True)
        return sheets_concat



# lasciate ogni speranza, voi ch'entrate

import os
import typing

from g3core import G3Core

from .type_hinting import (
    SWDeviceDict,
    SWModuleDict,
    SWSystemDict
)
from ._sw_system_dict_wrapper import SWSystemDictWrapper
from ._logger import logger

if typing.TYPE_CHECKING:
    from g3core.st_parser import STStruct, STFunctionBlock, STVariable


class SWSystemDictG3CoreUpdater:
    ROUTE_ELEMENT_FB = {
        'pointmachine': 'fRoute_PM',
        'detector': 'fRoute_detector',
        'crossing': 'fRoute_crossing'
        }
    ROUTE_PM_POSITIONS = {
        'l': 'ROUTE_DIRECTION_LEFT',
        'r': 'ROUTE_DIRECTION_RIGHT',
        'default': 'ROUTE_DIRECTION_NONE'
        }

    def __init__(
        self,
        system: SWSystemDict,
        g3core: G3Core,
    ) -> None:
        self.wrapper = SWSystemDictWrapper(system)
        self.g3core = g3core

    def _update_device_connector_control_var(
        self,
        var: 'STVariable',
        zone_name: str,
        device_type: str,
        device_data: SWDeviceDict,
        connectors: dict[str, str | typing.Any]
    ) -> None:
        match var.name:
            case 'dev':
                var.value = device_data['general']['varname']
            case 'comm':
                connected_commfunc = connectors.get('comm')
                if connected_commfunc:
                    var.value = connected_commfunc
            case 'system':
                var.value = 'System'
            case 'zone':
                zone_data = self.wrapper.find_device(
                    zone_name, 'Zone', zone_name
                    )
                if zone_data is None:
                    logger.warning(
                        'Could not find data for Zone "%s". '
                        'Connector will be set to NULL.',
                        zone_name
                        )
                    var.value = None
                    return
                zone_varname = zone_data['general']['varname']
                var.value = zone_varname
            case 'fromCabinet':
                if 'fromCabinet' not in connectors:
                    logger.warning(
                        '%s "%s": "fromCabinet" connector is empty. '
                        'Connector will be set to NULL.',
                        device_type.capitalize(),
                        device_data["general"]["name"]
                        )
                    var.value = None
                    return
                connected_cabinet_data = self.wrapper.find_device(
                    zone_name,
                    'Cabinet',
                    connectors['fromCabinet']
                    )
                if not connected_cabinet_data:
                    logger.warning(
                        '%s "%s": "fromCabinet" connector is empty. '
                        'Connector will be set to NULL.',
                        device_type.capitalize(),
                        device_data["general"]["name"]
                        )
                    var.value = None
                    return
                connected_cabinet_varname = (
                    connected_cabinet_data['general']['varname']
                    )
                device_type = device_type.lower().capitalize()
                var.value = f'{connected_cabinet_varname}.to{device_type}'
            case _:
                if var.name not in connectors:
                    logger.warning(
                        '%s "%s": "%s" connector is added automatically '
                        'without a reference to any device.',
                        device_type.capitalize(),
                        device_data["general"]["name"],
                        var.name
                        )
                    var.value = None
                    return
                connected_device_name = connectors[var.name]
                if not connected_device_name:
                    logger.warning(
                        '%s "%s": "%s" connector is empty.',
                        device_type.capitalize(),
                        device_data["general"]["name"],
                        var.name
                        )
                    var.value = None
                    return
                if any(
                    req_type in device_type.lower()
                    for req_type in ['vecom', 'spie', 'drr']
                ) and var.name == 'controller':
                    req_type_tmp = device_type
                    req_type = device_type.removesuffix('Loop')
                    if(req_type == req_type_tmp):
                        req_type = device_type.removesuffix('Transceiver')
                    connected_device_type = f'{req_type}Controller'
                else:
                    name = f'{var.name[0].capitalize()}{var.name[1:]}'
                    connected_device_type = name
                connected_device_data = self.wrapper.find_device(
                    zone_name,
                    connected_device_type,
                    connected_device_name
                    )
                if not connected_device_data:
                    device_connector_path = (
                        f'{device_type}/{device_data["general"]["name"]}/'
                        f'control/connector/{var.name}'
                        )
                    logger.warning(
                        f'Could not find data for {connected_device_type} '
                        f'"{connected_device_name}" connected at '
                        f'"{device_connector_path}". '
                        f'Connector will be set to NULL.'
                        )
                    var.value = None
                else:
                    connected_device_varname = (
                        connected_device_data['general']['varname']
                        )
                    var.value = connected_device_varname

    def update_device_connectors_control(
        self,
        zone_name: str,
        device_type: str,
        device_data: SWDeviceDict,
        fb: 'STFunctionBlock'
    ) -> None:
        connectors = device_data['control'].get('connector', {})
        connectors_filtered = {}
        for conn_name, conn_value in connectors.items():
            if any(conn_name == var.name for var in fb.var_in_out):
                connectors_filtered[conn_name] = conn_value
            else:
                logger.warning(
                    '%s "%s": "%s" function block does not have a "%s" '
                    'IN_OUT variable. The connector will be removed. ',
                    device_type.capitalize(),
                    device_data["general"]["name"],
                    fb.name,
                    conn_name
                    )
        connectors = connectors_filtered
        for var in fb.var_in_out:
            self._update_device_connector_control_var(
                var, zone_name, device_type, device_data, connectors
                )
            connectors[var.name] = var.value
        device_data['control']['connector'] = connectors

    def update_device_connectors_test(
        self,
        zone_name: str,
        device_type: str,
        device_data: SWDeviceDict,
        fb: 'STFunctionBlock'
    ) -> None:
        if 'connector' not in device_data['test']:
            device_data['test']['connector'] = {}
        for var in fb.var_in_out:
            match var.name:
                case 'dev':
                    var.value = f"Test_{device_data['general']['varname']}"
                case 'comm':
                    control_commfunc = (
                        device_data['control']['connector'].get('comm')
                        )
                    if control_commfunc:
                        var.value = f"Test_{control_commfunc}"
                case 'system':
                    if device_type == 'System':
                        var.value = 'System'
                    else:
                        var.value = 'TestSystem'
                case _:
                    if var.name not in device_data['test']['connector']:
                        logger.warning(
                            'test/connector/%s: Reference to connected device '
                            'is empty.', var.name
                            )
                        continue
                    control_device_data = self.wrapper.find_device(
                        zone_name,
                        var.name.capitalize(),
                        device_data['control']['connector'][var.name]
                        )
                    if not control_device_data:
                        raise ValueError  # do warning here
                    connected_device_varname = (
                        f"Test_{control_device_data['general']['varname']}"
                        )
                    var.value = connected_device_varname
            device_data['test']['connector'][var.name] = var.value

    def update_device_config_control(
        self,
        zone_name: str,
        device_data: SWDeviceDict,
        config_struct: 'STStruct'
    ) -> None:
        if 'config' not in device_data['control']:
            device_data['control']['config'] = {}
        for var in config_struct.members:
            match var.name:
                case 'taskName':
                    value = self.wrapper.get_zone_taskname(zone_name)
                    var.value = value
                case 'varName':
                    value = self.wrapper.get_device_varname(device_data)
                    var.value = value
                # this is a hotfix to handle nested config data in Comm:SHV
                case 'parentBroker':
                    value = device_data['control']['config'][var.name]
                    assert isinstance(value, dict)
                    assert list(value.keys()) == ['connection']
                    assert isinstance(value['connection'], dict)
                    assert list(value['connection'].keys()) == [
                        'serverAddress', 'serverPort'
                        ]
                    server_addr = value['connection']['serverAddress']
                    if not server_addr or not isinstance(server_addr, str):
                        server_addr = "nirvana.elektroline.cz"
                        logger.warning(
                            '"parentBroker/connection/serverAddress": cannot '
                            'format value "%s" to data type "string". '
                            'Default value "%s" will be used instead.',
                            server_addr, server_addr
                            )
                    value['connection']['serverAddress'] = server_addr
                    server_port = value['connection']['serverPort']
                    try:
                        server_port = int(server_port)
                        if server_port <= 0:
                            raise ValueError("Value must be a positive number")
                    except ValueError as err:
                        server_port = 3756
                        logger.warning(
                            '"parentBroker/connection/serverPort": cannot '
                            'format value "%s" to data type "int" (%s). '
                            'Default value "%s" will be used instead.',
                            server_port, err, server_port
                            )
                    value['connection']['serverPort'] = server_port
                    var.value = value
                case _:
                    if var.name in device_data['control']['config']:
                        value = device_data['control']['config'][var.name]
                        var.value = value
            device_data['control']['config'][var.name] = var.to_py()

    def update_device_config_test(
        self,
        zone_name: str,  # for method signature consistency
        device_data: SWDeviceDict,
        config_struct: 'STStruct'
    ) -> None:
        if 'config' not in device_data['test']:
            device_data['test']['config'] = {}
        for var in config_struct.members:
            if var.name in device_data['test']['config']:
                var.value = device_data['test']['config'][var.name]
            elif var.value is None:
                logger.warning(
                    'test/config/%s: config parameter is empty.', var.name
                    )
            device_data['test']['config'][var.name] = var.to_py()

    def prefill_device_iomap_test(
        self, device_data: SWDeviceDict, fb: 'STFunctionBlock'
    ) -> None:
        if 'test' not in device_data:
            device_data['test'] = {}
        if 'iomap' not in device_data['test']:
            device_data['test']['iomap'] = {}
        for var in fb.var_input + fb.var_out:
            if var.name.startswith('in'):
                varname = f'out{var.name[2:]}'
            elif var.name.startswith('out'):
                varname = f'in{var.name[3:]}'
            else:
                logger.warning(
                    'Variable "%s" in function block "%s" has an unexpected '
                    'prefix (expected "in" or "out" prefix). Variable will be '
                    'added to the iomap as is.', var.name, fb.name
                    )
                varname = var.name
            device_data['test']['iomap'][var.name] = varname

    def update_device(
        self,
        zone_name: str,
        device_type: str,
        device_domain: typing.Literal['control', 'test'],
        device_data: SWDeviceDict
    ) -> None:
        logger.info(
            'Looking for %s "%s" %s function block type', device_type,
            self.wrapper.get_device_name(device_data), device_domain
            )
        fb_name = device_data[device_domain]['function']
        # validate if the function block name is specified
        if not fb_name:
            logger.warning(
                '%s function block type of %s "%s" is not specified. '
                'Device %s config data cannot be validated.',
                device_domain.capitalize(), device_type,
                self.wrapper.get_device_name(device_data), device_domain
                )
            return
        # add the test device varname to the general data
        if device_domain == 'test':
            varname_test = f"Test_{device_data['general']['varname']}"
            device_data['general']['varname_test'] = varname_test
        # get fb data corresponding to the specified fb name
        try:
            fb_path, fb = self.g3core.find_function_block(fb_name)
        except (KeyError, ValueError):
            logger.warning(
                '%s function block type of %s "%s" was not found in G3 Core '
                'files. Device %s config data cannot be validated.',
                device_domain.capitalize(), device_type,
                self.wrapper.get_device_name(device_data), device_domain
                )
            return
        # prefill test iomap keys with VAR_IN and VAR_OUT fb members
        if device_domain == 'test':
            logger.info(
                'Prefilling test VAR_IN and VAR_OUT keys for %s "%s"',
                device_type, self.wrapper.get_device_name(device_data)
                )
            self.prefill_device_iomap_test(device_data, fb)
        # update function block VAR_IN_OUT variables ("connectors")
        updader = getattr(self, f'update_device_connectors_{device_domain}')
        updader(zone_name, device_type, device_data, fb)
        # get the "dev" VAR_IN variable
        vars_dev = [var for var in fb.var_in_out if var.name == 'dev']
        if not vars_dev:
            logger.info(
                '"dev" member was not found in %s function block "%s". '
                'Device config data will be discarded.', device_domain, fb.name
                )
            device_data[device_domain]['config'] = {}
            device_data[device_domain]['devType'] = ''
            return
        var_dev = vars_dev[0]
        # get the "dev" struct data (raises an error if unsuccessful)
        dev_struct_path, dev_struct = self.g3core.find_struct(
            var_dev.dtype,  # the name of the struct
            search_here_first=f'{os.path.dirname(fb_path)}/Types.typ'
            )
        # update "dev" struct name in the device control data
        device_data[device_domain]['devType'] = dev_struct.name
        # get the "config" member (also a struct) of the "dev" struct
        try:
            memb_config = [
                memb for memb in dev_struct.members if memb.name == 'config'
                ][0]
        except IndexError:
            logger.info(
                '"config" member was not found in %s function block "%s". '
                'Device config data will be discarded.', device_domain, fb.name
                )
            device_data[device_domain]['config'] = {}
            return
        # get the "config" struct data (raises an error if unsuccessful)
        config_struct_path, config_struct = self.g3core.find_struct(
            memb_config.dtype, search_here_first=dev_struct_path
            )
        updader = getattr(self, f'update_device_config_{device_domain}')
        updader(zone_name, device_data, config_struct)

    def update_route_layout(self, route_data: SWDeviceDict) -> None:
        # remove and store the "route_elements" key value
        layout_str: str = route_data['general'].pop('route_elements')
        if not layout_str:
            route_data['general']['layout'] = []
            route_data['general']['length'] = 0
            return
        # define standard element length
        pm_len = 4
        det_len = 8
        cross_len = 2
        space_len = 2
        element_startoffset = 2
        route_len = 0
        # process the "route_elements" key value
        layout = []
        for element in layout_str.split(','):
            element = element.strip(' ')
            kwargs = {}  # additional element-specific key-value data pairs
            # get the element type and params
            if '@' in element:
                position, element_name = element.split('@')
                position = self.ROUTE_PM_POSITIONS.get(
                    position.lower(), self.ROUTE_PM_POSITIONS['default']
                    )
                element_type = 'pointmachine'
                element_len = pm_len
                kwargs['position'] = position
            elif '#' in element:
                element_name = f"CROSSING{element.lstrip('#')}"
                element_type = 'crossing'
                element_len = cross_len
            else:
                element_name = element
                element_type = 'detector'
                element_len = det_len
            # create and save the element data dict
            element_endoffset = element_startoffset + element_len
            element_data = {
                'name': element_name,
                'type': element_type,
                'function': self.ROUTE_ELEMENT_FB[element_type],
                'startoffset': element_startoffset,
                'endoffset': element_endoffset,
                }
            element_data.update(kwargs)
            layout.append(element_data)
            element_startoffset = element_endoffset + space_len
        # save the created list of layout elements
        route_len = element_startoffset
        route_data['general']['layout'] = layout
        route_data['general']['length'] = route_len

    def _validate_module(
        self,
        zone_name: str,
        module_name: str,
        module_data: SWModuleDict,
        ignore: list[str] | None = None
    ) -> None:
        if ignore and module_name in ignore:
            logger.debug(
                'Module data validation for "%s/%s" is ignored',
                zone_name, module_name
                )
            return
        logger.info(
            'Validating module data for "%s/%s"', zone_name, module_name
            )
        for device_name, device_data in module_data.items():
            logger.info(
                'Validating control device data for "%s/%s/%s"',
                zone_name, module_name, device_name
                )
            self.update_device(
                zone_name, module_name, 'control', device_data
                )
            logger.info(
                'Validating test device data for "%s/%s/%s"',
                zone_name, module_name, device_name
                )
            self.update_device(
                zone_name, module_name, 'test', device_data
                )
            if module_name == 'Route':
                logger.info(
                    'Updating route "%s/%s/%s" layout data',
                    zone_name, module_name, device_name
                    )
                self.update_route_layout(device_data)
            logger.debug(
                'Validating children device data for device "%s/%s/%s"',
                zone_name, module_name, device_name
                )
            children = device_data.setdefault('children', {})
            if children:
                for child_module_name, child_module_data in children.items():
                    self._validate_module(
                        zone_name, child_module_name, child_module_data, ignore
                        )

    def validate(self) -> None:
        logger.info(
            'Validating system data for "%s"',
            self.wrapper.get_system_name()
            )
        for zone_name, zone_data in self.wrapper.system_dict.items():
            logger.info('Validating zone data for "%s"', zone_name)
            for module_name, module_data in zone_data.items():
                self._validate_module(
                    zone_name,
                    module_name,
                    module_data,
                    ignore=['Project', 'System']
                    )



import pandas as pd
import logging
import typing
import warnings

from ..utils.gdrive import get_table_data_from_gdrive


logger = logging.getLogger('g3tables.visualization_table')


DEVICE_TYPES = {
    "System": "System_G3",
    "SystemSafety": "SystemSafety_G3",
    "Zone": "Zone_G3",
    "Gate": "Gate_G3",
    "Route": "Route_G3",
    "Zone.Track": "Track",
    "Detector.TrackCircuit": "TC_G3",
    "Detector.Pantograph": "PD_G3",
    "Detector.MassDetector": "MD_G3",
    "Detector.UltrasonicSensor": "US_G3",
    "Detector.VirtualDetector": "VD_G3",
    "PointMachine.PME": "PME_G3",
    "PointMachine.PMM": "PMM_G3",
    "Signal": "Signal_G3",
    "Signal.Symbol": "SignalSymbol_G3",
    "DoorDisplay7": "DoorDisplay7_G3",
    "Requestor.Digital": "RequestorDigital",
    "Requestor.RoutingTable": "RoutingTable_G3",
    "Requestor.Vecom": "VecomController_G3",
    "Requestor.Vecom.Loop": "VecomLoop_G3",
    "Requestor.DRR": "DRRController_G3",
    "Requestor.DRR.Transceiver": "DRRTransceiver_G3",
    "Requestor.SPIE": "SPIEController_G3",
    "Requestor.SPIE.Loop": "SPIELoop_G3",
    "Requestor.Vetra": "Vetra_G3",
    "Requestor.AWA": "AWA_G3",
    "Cabinet": "Cabinet_G3",
    "Cabinet.UPS": "CabinetUps_G3",
    "Cabinet.RFID": "CabinetRFID_G3",
    "Cabinet.Fuse": "CabinetFuse_G3",
    "Cabinet.Convertor": "CabinetConvertor_G3",
    "Cabinet.MonitoringModule": "CabinetModule_G3",
    "Heating": "Heating_G3",
    "Heating.Contactor": "HeatingContactor_G3",
    "Heating.Contactor.Rod": "HeatingRod_G3",
    "Matrix": "MPS_G3",
    "GPIO": "GPIO_G3" 
}


class SHVVarDict(typing.TypedDict):
    name: str
    is_blacklisted: bool
    is_deleted: bool


class SHVProjectTypeDict(typing.TypedDict):
    type: str
    restricted_to: str | None
    vars: dict[str, SHVVarDict]
    type_variants: typing.NotRequired[dict[str, dict]]


GateTypeVariantTableDict: typing.TypeAlias = dict[str, tuple[list[str], int]]

HeatingTypeVariantTableDict: typing.TypeAlias = dict[str, list[str]]

RequestorTypeVariantTableDict: typing.TypeAlias = dict[str, list[str]]


class VisualizationTable:
    """
    `VisualizationTable` is a representation of the G3 Project Visualization
    table. It allows access to the data of the 'Parameters', 'Gate table',
    'Heating table', and 'Requestor table' sheets in the form of a
    `pandas.DataFrame` dictionary.

    Note that the "Title", "Identification", and "Worksheet explanation" sheets
    are not represented within the dictionary.
    """
    NAME_PATTERN = 'project_visualization'
    NAME_PATTERN_ASSESSMENT = 'system_hmi_visualization'
    """
    The pattern that the name of the SW Definition table must match to be
    automatically recognized in a Google Drive folder.
    """
    EXCLUDED_SHEETS = [
        'Title', 'Identification', 'Worksheet explanation', 'Users'
        ]
    """
    Sheets that are not represented in the data dictionary of
    a `SWDefinitionTable` instance.
    """

    def __init__(
        self,
        parameters_sheet_data: pd.DataFrame,
        gate_table_sheet_data: pd.DataFrame,
        heating_table_sheet_data: pd.DataFrame,
        requestor_table_sheet_data: pd.DataFrame
    ) -> None:
        """
        Standard initializer of a `VisualizationTable` instance. Although it
        can be used directly, alternative class constructors `from_local` and
        `from_gdrive` are adviced instead.

        Args:
            data (dict[str, pd.DataFrame]): a dict, where keys are\
            sheets names and values are the `pandas.DataFrame` objects\
            representing the sheet data.
        """
        self._params_data = parameters_sheet_data
        self._gate_data = gate_table_sheet_data
        self._heating_data = heating_table_sheet_data
        self._req_data = requestor_table_sheet_data

    @classmethod
    def is_name_valid(cls, name: str) -> bool:
        """
        Check if the provided name matches the Visualization table
        name pattern.

        Args:
            name (str): The name to validate.

        Returns:
            bool: True if the name is valid, False otherwise.
        """
        name_formatted = name.strip().replace(' ', '_').casefold()
        return ( (cls.NAME_PATTERN in name_formatted) or (cls.NAME_PATTERN_ASSESSMENT in name_formatted) )

    @classmethod
    def from_local(cls, path: str) -> typing.Self:
        """
        Create a `VisualizationTable` instance from a local Excel file.

        Args:
            local_path (str): The local file path to the Excel file containing\
                the Visualization table data.

        Returns:
            Self: An instance of `VisualizationTable` initialized\
                with data from the specified local Excel file.
        """
        logger.info('Reading Visualization table from local path: %s', path)
        xls = pd.ExcelFile(path)
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning)
            parameters_sheet = pd.read_excel(
                xls, header=2, sheet_name='Parameters'
                )
            gate_table_sheet = pd.read_excel(
                xls, header=2, sheet_name='Gate table'
                )
            heating_table_sheet = pd.read_excel(
                xls, header=2, sheet_name='Heating table'
                )
            requestor_table_sheet = pd.read_excel(
                xls, header=2, sheet_name='Requestor table'
                )
        params_formatter = VisualizationTableFormatter.format_params_sheet
        table_formatter = VisualizationTableFormatter.format_table_sheet
        return cls(
            parameters_sheet_data=params_formatter(parameters_sheet),
            gate_table_sheet_data=table_formatter(gate_table_sheet),
            heating_table_sheet_data=table_formatter(heating_table_sheet),
            requestor_table_sheet_data=table_formatter(requestor_table_sheet),
            )

    @classmethod
    def from_gdrive(cls, gdrive_table_name: str) -> typing.Self:
        """
        Create a `VisualizationTable` instance from a Google Drive file.

        Args:
            gdrive_table_name (str): The name of the table in Google Drive\
                containing the Visualization table.

        Returns:
            SWDefinitionTable: An instance of `VisualizationTable` initialized\
                with data from the specified Google Drive file.
        """
        logger.info(
            'Reading Visualization table from Google Drive: %s',
            gdrive_table_name
            )
        sheets_all = get_table_data_from_gdrive(
            gdrive_table_name, header_row=3, exclude_sheets=cls.EXCLUDED_SHEETS
            )
        parameters_sheet = sheets_all['Parameters']
        gate_table_sheet = sheets_all['Gate table']
        heating_table_sheet = sheets_all['Heating table']
        requestor_table_sheet = sheets_all['Requestor table']
        params_formatter = VisualizationTableFormatter.format_params_sheet
        table_formatter = VisualizationTableFormatter.format_table_sheet
        return cls(
            parameters_sheet_data=params_formatter(parameters_sheet),
            gate_table_sheet_data=table_formatter(gate_table_sheet),
            heating_table_sheet_data=table_formatter(heating_table_sheet),
            requestor_table_sheet_data=table_formatter(requestor_table_sheet),
            )

    @classmethod
    def load(cls, path: str) -> typing.Self:
        """
        Load a Visualization table from a local file or Google Drive.

        Args:
            path (str): Path to the Visualization table file. If the path ends\
                with '.xlsx', the table is loaded from a local file.\
                Otherwise, the table is loaded from Google Drive.

        Returns:
            typing.Self: An instance of `VisualizationTable` initialized with\
                data from the specified file.
        """
        if path.endswith('.xlsx'):
            return cls.from_local(path)
        return cls.from_gdrive(path)

    @property
    def params_data(self) -> pd.DataFrame:
        """'Parameters' sheet data of the Visualization table."""
        return self._params_data

    @property
    def gate_table_data(self) -> pd.DataFrame:
        """'Gate table' sheet data of the Visualization table."""
        return self._gate_data

    @property
    def heating_table_data(self) -> pd.DataFrame:
        """'Heating table' sheet data of the Visualization table."""
        return self._heating_data

    @property
    def requestor_table_data(self) -> pd.DataFrame:
        """'Requestor table' sheet data of the Visualization table."""
        return self._req_data

    @staticmethod
    def _new_shv_project_type_dict(
        module_type: str, restricted_to: str | None
    ) -> SHVProjectTypeDict:
        return {
            'type': module_type,
            'restricted_to': restricted_to,
            'vars': {}
            }

    @staticmethod
    def _new_shv_var_dict(
        shv_var_name: str, is_blacklisted: bool, is_deleted: bool
    ) -> SHVVarDict:
        return {
            'name': shv_var_name,
            'is_blacklisted': is_blacklisted,
            'is_deleted': is_deleted
            }

    @property
    def project_types(self) -> dict[str, SHVProjectTypeDict]:
        """Dictionary of project types and their SHV variables."""
        project_types: dict[str, SHVProjectTypeDict] = {}
        for _, row in self._params_data.iterrows():
            # create or fetch existing project_type_data dictionary
            project_type = row['Project type']
            project_subtype = row['Project subtype']
            if project_subtype:
                module_type, restricted_to = project_subtype, project_type
            else:
                module_type, restricted_to = project_type, None
            project_type_data = project_types.setdefault(
                module_type,
                self._new_shv_project_type_dict(module_type, restricted_to)
                )
            # create shv variable data dictionary
            if not (shv_var := row['SHV variable']):
                continue
            project_type_data['vars'][shv_var] = self._new_shv_var_dict(
                shv_var_name=shv_var,
                is_blacklisted=row['Available only in SHVspy'],
                is_deleted=row['x']
                )
        return project_types

    def _get_table_data(
        self, data: pd.DataFrame, start_idx: int
    ) -> dict[str, dict]:
        columns = VisualizationTableFormatter.get_sorted_columns(
            table_data=data.iloc[:, start_idx:]
            )
        tables: dict[str, dict] = {}
        i = 0
        for _, row in data.iterrows():
            type_ = row['Project type']
            type_variant = row['SHV visu type variant'] or 'default'
            table = {type_variant: columns[int(i)]}
            tables.setdefault(type_, {}).update(table)
            i += 1
        return tables

    @property
    def gate_tables(self) -> dict[str, GateTypeVariantTableDict]:
        """Dictionary of gate tables and their columns."""
        tables = self._get_table_data(self._gate_data, start_idx=4)
        for _, row in self._gate_data.iterrows():
            type_ = row['Project type']
            type_variant = row['SHV visu type variant'] or 'default'
            row_count = int(row['Row count'])
            columns = tables[type_][type_variant]
            tables[type_][type_variant] = (columns, row_count)
        return tables

    @property
    def heating_tables(self) -> dict[str, HeatingTypeVariantTableDict]:
        """Dictionary of heating tables and their columns."""
        return self._get_table_data(self._heating_data, start_idx=3)

    @property
    def requestor_tables(self) -> dict[str, RequestorTypeVariantTableDict]:
        """Dictionary of requestor tables and their columns."""
        return self._get_table_data(self._req_data, start_idx=3)


class VisualizationTableFormatter:

    @staticmethod
    def fill_in_device_types(device):
        if device in DEVICE_TYPES:
            return DEVICE_TYPES[device]
        logger.warning(
            f'Could not retrieve SHV Device type for module "{device}".'
            )
        return ''

    @staticmethod
    def format_params_sheet(params_sheet: pd.DataFrame) -> pd.DataFrame:
        # drop rows where the 'Module (device)' column value is empty
        params_sheet = params_sheet.fillna('').astype(str, copy=True)
        params_sheet_mask = (params_sheet['Module (device)'] != '')
        params_sheet = params_sheet[params_sheet_mask]
        # create the 'Project type' column
        formatter = VisualizationTableFormatter.fill_in_device_types
        params_sheet.insert(
            loc=1,
            column='Project type',
            value=params_sheet['Module (device)'].apply(formatter)
            )
        # format the 'x' column to bool
        params_sheet.loc[:, 'x'] = params_sheet['x'].apply(
            lambda cell_val: bool(cell_val)
        )
        # format the 'Available only in SHVspy' column to bool
        col_name = 'Available only in SHVspy'
        params_sheet.loc[:, col_name] = params_sheet[col_name].apply(
            lambda cell_val: cell_val == 'TRUE'
        )
        return params_sheet

    @staticmethod
    def format_table_sheet(table_sheet: pd.DataFrame) -> pd.DataFrame:
        # drop rows where the 'Module (device)' column value is empty
        table_sheet = table_sheet.fillna('').astype(str, copy=True)
        empty_module_mask = table_sheet['Module (device)'].str.fullmatch('')
        table_sheet = table_sheet[~empty_module_mask]
        # create the 'Project type' column
        formatter = VisualizationTableFormatter.fill_in_device_types
        table_sheet.insert(
            loc=0,
            column='Project type',
            value=table_sheet['Module (device)'].apply(formatter)
            )
        return table_sheet

    @staticmethod
    def get_sorted_columns(table_data: pd.DataFrame) -> list[list[str]]:

        def sort_columns_by_row_values(row):
            sorted_values = sorted(
                [
                    (col, val) for col, val in row.items()
                    if val and int(val) != 0
                ],
                key=lambda x: x[1]
                )
            return [col for col, _ in sorted_values]

        data_sorted = table_data.apply(sort_columns_by_row_values, axis=1)
        return data_sorted.values.tolist()



import typing

from ..visualization_table import SHVProjectTypeDict
from ..visualization_table import SHVVarDict  # noqa: F401


class HWModuleDict(typing.TypedDict):
    type: str
    cabinet: str | None
    name_suffix: str


class HWIsleDict(typing.TypedDict):
    head: HWModuleDict
    tail: list[HWModuleDict]


class SWDeviceConfigDict(typing.TypedDict):
    config: dict[str, typing.Any]
    function: str
    connector: dict[str, str | typing.Any]
    devType: str
    iomap: typing.NotRequired[dict[str, str]]


class SWDeviceVisuDict(typing.TypedDict):
    SHVDeviceType: str
    SHVDeviceTypeVariant: typing.NotRequired[str]


class SWDeviceDict(typing.TypedDict):
    general: dict[str, typing.Any]
    control: SWDeviceConfigDict
    test: SWDeviceConfigDict
    visu: SWDeviceVisuDict
    children: dict[str, 'SWModuleDict']


SWModuleDict: typing.TypeAlias = dict[str, SWDeviceDict]
SWZoneDict: typing.TypeAlias = dict[str, SWModuleDict]
SWSystemDict: typing.TypeAlias = dict[str, SWZoneDict]


class SystemDict(typing.TypedDict):
    Hardware: list[HWIsleDict]
    Software: SWSystemDict
    Visu: dict[str, SHVProjectTypeDict]



import typing

from .type_hinting import (
    SWDeviceDict,
    SWModuleDict,
    SWSystemDict
)


class SWSystemDictWrapper:
    ZONE_TASKNAME_MAXLEN = 10

    def __init__(self, system: SWSystemDict) -> None:
        self.system_dict = system

    @property
    def common_data(self) -> dict:
        return {
            module: data
            for module, data in self.system_dict.get('Common', {}).items()
            if module not in ['Project']
            }

    @property
    def comm_data(self) -> dict:
        return {
            module: data
            for module, data in self.system_dict.get('Common', {}).items()
            if module not in ['Project', 'System', 'SystemSafety']
            }

    @property
    def redmine_data(self) -> dict[str, str]:
        for device_data in self.system_dict['Common']['Project'].values():
            return device_data['redmine']  # type: ignore
        return {}

    @property
    def zone_names(self) -> list[str]:
        return [
            name for name in list(self.system_dict.keys()) if name != 'Common'
            ]

    @property
    def zone_tasknames(self) -> list[str]:
        return [
            self.get_zone_taskname(name) for name in self.zone_names
            ]

    def get_zone_taskname(self, zone_name: str) -> str:
        if len(self.system_dict.keys()) < 3:  # not a multizone
            taskname = "Zone"
        else:
            zone_data = self.system_dict[zone_name]['Zone'][zone_name]
            varname = zone_data['general']['name']
            taskname = varname[:self.ZONE_TASKNAME_MAXLEN].strip('_')
        return taskname

    def has_system_safety(self) -> bool:
        return bool(self.system_dict['Common'].get('SystemSafety'))

    def has_shv(self) -> bool:
        return bool(self.system_dict['Common'].get('SHV'))

    def has_elesys(self) -> bool:
        return bool(self.system_dict['Common'].get('Elesys'))

    def get_project_name(self) -> str:
        for key in self.system_dict['Common']['Project'].keys():
            return key
        raise KeyError('Project data is empty.')

    def get_system_name(self) -> str:
        for key in self.system_dict['Common']['System'].keys():
            return key
        raise KeyError('System data is empty.')

    def get_project_libraries(self) -> list[str]:
        project_name = self.get_project_name()
        project_data = self.system_dict['Common']['Project'][project_name]
        libraries = project_data['general'].get('specificLibraries', '')
        assert isinstance(libraries, str)
        return [lib.strip() for lib in libraries.split(',')]

    @staticmethod
    def is_device_safety(device_data: SWDeviceDict) -> bool:
        func = device_data.get('control', {}).get('function', '').lower()
        return ('safety' in func) and ('non' not in func)

    @staticmethod
    def get_device_name(device_data: SWDeviceDict) -> str:
        return device_data['general']['name']

    @staticmethod
    def get_device_varname(device_data: SWDeviceDict) -> str:
        return device_data['general']['varname']

    @staticmethod
    def get_device_control_fb(device_data: SWDeviceDict) -> str | None:
        return device_data['control'].get('function')

    @staticmethod
    def get_device_test_fb(device_data: SWDeviceDict) -> str | None:
        return device_data['test'].get('function')

    @staticmethod
    def get_connected_device_type(connector_key: str) -> str:
        return connector_key.split('/').pop().capitalize()

    def _find_device(
        self,
        module_data: dict[str, SWModuleDict],
        connected_device_type: str,
        connected_device_name: str
    ) -> SWDeviceDict | None:
        try:
            devices = module_data[connected_device_type]
            if connected_device_name not in devices:
                return None
            return devices[connected_device_name]
        except KeyError:
            for devices in module_data.values():
                for device_data in devices.values():
                    device = self._find_device(
                        device_data.get('children', {}),
                        connected_device_type,
                        connected_device_name
                        )
                    if device is not None:
                        return device
            return None

    def find_device(
        self,
        zone_name: str,
        connected_device_type: str,
        connected_device_name: str
    ) -> SWDeviceDict | None:
        device = self._find_device(
            self.system_dict[zone_name],
            connected_device_type,
            connected_device_name
            )
        if device is not None:
            return device
        device = self._find_device(
            self.system_dict['Common'],
            connected_device_type,
            connected_device_name
            )
        return device

    def _collect_device(
        self,
        device_type: str,
        device_name: str,
        device_data: SWDeviceDict,
        device_path_prefix: str,
    ) -> typing.Iterator[tuple[str, str, SWDeviceDict]]:
        """->parent_device_path, device_path, device_data"""
        device_path = f'{device_type}/{device_name}'
        if device_path_prefix:
            device_path = f'{device_path_prefix}/{device_path}'
        yield (device_path_prefix, device_path, device_data)
        children = device_data.get('children')
        if children:
            for ch_module_name, ch_module_data in children.items():
                yield from self._collect_module(
                    ch_module_name, ch_module_data, device_path
                    )

    def _collect_module(
        self,
        module_name: str,
        module_data: SWModuleDict,
        module_path_prefix: str,
    ) -> typing.Iterator[tuple[str, str, SWDeviceDict]]:
        """->parent_device_path, device_path, device_data"""
        for device_name, device_data in module_data.items():
            yield from self._collect_device(
                module_name,
                device_name,
                device_data,
                module_path_prefix,
                )

    def iter_devices_zone(
        self, zone_name: str
    ) -> typing.Iterator[tuple[str, str, SWDeviceDict]]:
        """->parent_device_path, device_path, device_data"""
        zone_data = self.system_dict[zone_name]
        for module_name, module_data in zone_data.items():
            yield from self._collect_module(module_name, module_data, '')

    def iter_devices_system(
        self
    ) -> typing.Iterator[tuple[str, str, str, SWDeviceDict]]:
        """->zone, parent_device_path, device_path, device_data"""
        for zone_name in self.system_dict:
            for parent_path, path, data in self.iter_devices_zone(zone_name):
                yield (zone_name, parent_path, path, data)

    def flatten(self) -> dict[str, SWDeviceDict]:
        return {
            f'{zone_name}/{device_path}': device_data
            for zone_name, _, device_path, device_data
            in self.iter_devices_system()
            }



import pandas as pd
import logging
import re
import typing
import warnings

from unidecode import unidecode

from ..utils import Interval, to_nested_dict
from ..utils.gdrive import get_table_data_from_gdrive


logger = logging.getLogger('g3tables.sw_definition_table')


class SWDefinitionTable:
    """
    `SWDefinitionTable` is a representation of the G3 Project SW Defintion
    table. It allows access to the data of the 'Common' sheet and software zone
    configuration sheets in the form of a `pandas.DataFrame` dictionary.

    Note that the "Title", "Identification", and "Worksheet explanation" sheets
    are not represented within  the dictionary.
    """
    NAME_PATTERN = 'project_sw_definition'
    NAME_PATTERN_ASSESSMENT = 'system_sw_definition'
    """
    The pattern that the name of the SW Definition table must match to be
    automatically recognized in a Google Drive folder.
    """
    EXCLUDED_SHEETS = ['Title', 'Identification', 'Worksheet explanation']
    """
    Sheets that are not represented in the data dictionary of
    a `SWDefinitionTable` instance.
    """

    def __init__(self, data: dict[str, pd.DataFrame]) -> None:
        """
        Standard initializer of a `SWDefinitionTable` instance. Although it
        can be used directly, alternative class constructors `from_local` and
        `from_gdrive` are adviced instead.

        Args:
            data (dict[str, pd.DataFrame]): a dict, where keys are\
            sheets names and values are the `pandas.DataFrame` objects\
            representing the sheet data.
        """
        self._data = data

    @classmethod
    def is_name_valid(cls, name: str) -> bool:
        """
        Check if the provided name matches the SW Definition table
        name pattern.

        Args:
            name (str): The name to validate.

        Returns:
            bool: True if the name is valid, False otherwise.
        """
        name_formatted = name.strip().replace(' ', '_').casefold()
        return ( (cls.NAME_PATTERN in name_formatted) or (cls.NAME_PATTERN_ASSESSMENT in name_formatted) )

    @classmethod
    def from_local(cls, path: str) -> typing.Self:
        """
        Create a `SWDefinitionTable` instance from a local Excel file.

        Args:
            local_path (str): The local file path to the Excel file containing\
                the SW Definition table.

        Returns:
            Self: An instance of `SWDefinitionTable` initialized\
                with data from the specified local Excel file.
        """
        logger.info('Reading SW Definition table from local path: %s', path)
        formatter = SWDefinitionSheetFormatter
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            sheets_all = pd.read_excel(path, sheet_name=None)
            sheets = formatter.remove_excluded_sheets(sheets_all)
        for sheet_name, sheet_data in sheets.items():
            sheets[sheet_name] = formatter.process_sheet_init(sheet_data)
        return cls(sheets)

    @classmethod
    def from_gdrive(cls, gdrive_table_name: str) -> typing.Self:
        """
        Create a `SWDefinitionTable` instance from a Google Drive file.

        Args:
            gdrive_table_name (str): The name of the table in Google Drive\
                containing the SW Definition table.

        Returns:
            SWDefinitionTable: An instance of `SWDefinitionTable` initialized\
                with data from the specified Google Drive file.
        """
        logger.info(
            'Reading SW Definition table from Google Drive: %s',
            gdrive_table_name
            )
        sheets_all = get_table_data_from_gdrive(
            gdrive_table_name,
            exclude_sheets=cls.EXCLUDED_SHEETS
            )
        formatter = SWDefinitionSheetFormatter
        sheets = formatter.remove_excluded_sheets(sheets_all)  # type: ignore
        for sheet_name, sheet_data in sheets.items():
            sheets[sheet_name] = formatter.process_sheet_init(sheet_data)
        return cls(sheets)

    @classmethod
    def load(cls, path: str) -> typing.Self:
        """
        Load a SW Definition table from a local file or Google Drive.

        Args:
            path (str): Path to the SW Definition table file. If the path ends\
                with '.xlsx', the table is loaded from a local file.\
                Otherwise, the table is loaded from Google Drive.

        Returns:
            typing.Self: An instance of `SWDefinitionTable` initialized with\
                data from the specified file.
        """
        if path.endswith('.xlsx'):
            return cls.from_local(path)
        return cls.from_gdrive(path)

    @property
    def sheet_names(self) -> list[str]:
        """Names of the sheets contained within the SW Definition table."""
        return [name for name in self._data.keys()]

    def get_sheet_data(self, sheet_name: str) -> pd.DataFrame:
        """
        Retrieves data for a specific sheet.

        Args:
            sheet_name (str): The name of the sheet to retrieve.

        Returns:
            pd.DataFrame: The data associated with the sheet.

        Raises:
            KeyError: If the specified sheet is not found in the data.
        """
        try:
            return self._data[sheet_name]
        except KeyError as err:
            err.add_note('Sheet was not found')
            raise err

    def get_system_sheets(self, system_name: str) -> dict[str, pd.DataFrame]:
        """
        Gets the sheets' data associated with a given system name. Note that
        the system name is not validated.

        Args:
            system_name (str): The name of the system for which to retrieve\
                sheets.

        Returns:
            dict[str, pd.DataFrame]: A dictionary of sheets belonging to\
                the given system.
        """
        sheets = {}
        for sheet_name, sheet_data in self._data.items():
            if sheet_name != 'Common':
                system_connector = self.get_connected_system(sheet_name)
                if system_connector != system_name:
                    continue
            sheets[sheet_name] = sheet_data
        if len(sheets) < 2:
            logger.error(
                'No zone sheets were found for system "%s". Aborting',
                system_name
                )
            exit(1)
        return sheets

    def get_connected_devices(
        self, zone_name: str, connector: str
    ) -> list[str]:
        """
        Retrieve connected device names based on the zone name and the type of
        connector. For instance, `get_connected_devices('Z01', 'gate')` returns
        a list of gate devices, which are connected to any other device via
        "control/connector/gate".

        Args:
            zone_name (str): Name of the zone sheet to look up.
            connector (str): Type of connector to find (gate, cabinet, etc.).

        Returns:
            list[str]: List of connected devices found.
        """
        logger.debug(
            'Getting connected device names for zone "%s" and connector '
            '"control/connector/%s"', zone_name, connector
            )
        zone_data = self.get_sheet_data(zone_name)
        conn_rows = zone_data.xs(f'control/connector/{connector}', level='Key')
        values = []
        column_index = 1
        while True:
            if column_index not in zone_data:
                break
            column_values = conn_rows[column_index]
            for cell_value in column_values.to_list():
                if not cell_value:
                    continue
                if cell_value not in values:
                    values.append(cell_value)
            column_index += 1
        return values

    def get_connected_system(self, zone_name: str) -> str:
        """
        Get the name of the system to which the specified zone belonds to.

        Args:
            zone_name (str): Name of the zone to query for system connector.

        Returns:
            str: The connected system connector.

        Raises:
            AttributeError: If no system connector is found or
            if it's ambiguous.
        """
        system_connector = self.get_connected_devices(zone_name, 'system')
        logger.debug(
            'Retrieved the system connector "%s" for zone "%s".',
            system_connector, zone_name
            )
        if not system_connector:
            raise AttributeError(
                f'Sheet "{zone_name}" does not have a system connector.'
                )
        if len(system_connector) > 1:
            connectors = ", ".join(system_connector)
            raise AttributeError(
                f'Sheet "{zone_name}" system connector is ambiguous '
                f'({len(system_connector)} connectors found: {connectors}).'
                )
        return system_connector[-1]

    def get_connected_commfunc_brc(self, zone_name: str) -> str | None:
        """
        Get the name of the BRC comm function connector if a TrackCircuit
        detector is present in the specified zone. If no TrackCircuit detector
        is found, the method returns None.

        Args:
            zone_name (str): Name of the zone to query for\
                the BRC comm function.

        Returns:
            str | None: The name of the BRC comm function connector if a\
                TrackCircuit detector is present, otherwise None.
        """
        logger.debug(
            'Looking for a BRC comm function within the "Common" sheet.'
            )
        common_data = self.get_sheet_data('Common')
        commfunc_row = common_data.loc[('CommFunction', 'general/name'), 1:]
        commfunc_brc = None
        column_index = 1
        while True:
            if column_index not in commfunc_row:
                break
            cell_value = commfunc_row[column_index]
            if cell_value and 'brc' in str(cell_value).lower():
                commfunc_brc = cell_value
                break
            column_index += 1
        if not commfunc_brc:
            logger.debug(
                'No BRC comm function was found within the "Common" sheet.'
                )
            return None
        assert isinstance(commfunc_brc, str)
        logger.debug('Found a BRC comm function within the "Common" sheet.')
        logger.debug(
            'Looking for a TrackCircuit detector within zone "%s".', zone_name
            )
        zone_data = self.get_sheet_data(zone_name)
        detector_type_row = zone_data.loc[('Detector', 'general/type'), 1:].squeeze()
        assert isinstance(detector_type_row, pd.Series)
        column_index = 1
        while True:
            if column_index not in detector_type_row:
                break
            cell_value = detector_type_row[column_index]
            assert isinstance(cell_value, str)
            if not cell_value:
                continue
            if cell_value.lower() == 'trackcircuit':
                logger.debug(
                    'Found a TrackCircuit detector within zone "%s".',
                    zone_name
                    )
                return commfunc_brc
            column_index += 1
        logger.debug(
            'No TrackCircuit detector was found within zone "%s".', zone_name
            )
        return None

    def get_connected_commfuncs(self, zone_name: str) -> list[str]:
        """
        Get the names of the communication function block instances
        which are utilized within the specified zone.

        Args:
            zone_name (str): Name of the zone to query.

        Returns:
            list[str]: List of communication function connectors.
        """
        # get comm func connectors that are represented in the Zone sheets
        logger.debug(
            'Retrieving the comm function connectors for zone "%s".', zone_name
            )
        commfunc_connectors = self.get_connected_devices(zone_name, 'comm')
        logger.debug(
            'Retrieved the comm function connectors "%s" for zone "%s".',
            commfunc_connectors, zone_name
            )
        # evaluate whether to add the 'CommBRC' connector
        logger.debug(
            'Evaluating whether to add a BRC comm function connector '
            'for zone "%s".', zone_name
            )
        commfunc_brc = self.get_connected_commfunc_brc(zone_name)
        if commfunc_brc:
            logger.debug(
                'Adding the "%s" BRC comm function connector to the comm '
                'function connectors for zone "%s".', commfunc_brc, zone_name
                )
            commfunc_connectors.append(commfunc_brc)
        else:
            logger.debug(
                'No BRC comm function connector is added to the comm '
                'function connectors for zone "%s".', zone_name
                )
        return commfunc_connectors

    def get_cabinet_names(self, zone_name: str) -> list[str]:
        """
        Retrieve the names of cabinets present in the specified zone.

        Args:
            zone_name (str): Name of the zone to query for cabinet names.

        Returns:
            list[str]: List of the found cabinet names.
        """
        logger.debug('Retrieving cabinet names for zone "%s".', zone_name)
        zone_data = self.get_sheet_data(zone_name)
        conn_row = zone_data.loc[('Cabinet', 'general/name'), 1:].squeeze()
        assert isinstance(conn_row, pd.Series)
        values: list[str] = []
        column_index = 1
        while True:
            if column_index not in zone_data:
                break
            cell_value = conn_row[column_index]
            assert isinstance(cell_value, str)
            if cell_value and (cell_value not in values):
                values.append(cell_value)
            column_index += 1
        logger.debug(
            'Retrieved cabinet names "%s" for zone "%s"', values, zone_name
            )
        return values

    def to_dict(self, system_name: str) -> dict[str, dict]:
        """
        Converts the system sheets' data to a pure tree-like dictionary
        representation.

        Args:
            system_name (str): The name of the system for which to generate\
                the dictionary.

        Returns:
            dict[str, dict]: The dictionary representation of the system's\
                sheets' data.
        """
        # get unprocessed system sheets related to the system
        sheets = self.get_system_sheets(system_name)
        # process zone sheets and gather connected comm functions
        zone_sheets_converted: dict[str, dict] = {}
        commfunc_names = []
        for sheet_name, sheet_data in sheets.items():
            if sheet_name == 'Common':
                continue
            converter = SWDefinitionSheetToNestedDictConverter(sheet_data)
            commfunc_names.extend(self.get_connected_commfuncs(sheet_name))
            zone_sheets_converted[sheet_name] = converter.to_dict()
        # process the 'Common' sheet
        commom_data = sheets['Common']
        converter = SWDefinitionCommonSheetToNestedDictConverter(commom_data)
        common_sheet_converted = {
            'Common': converter.to_dict_common(system_name, commfunc_names)
            }
        # update the dict with the process 'Common' sheets with the processed
        # zone sheets (this way, the 'Common' key appears first in the dict)
        common_sheet_converted.update(zone_sheets_converted)
        # expand all the remaining slash-separated keys
        # to create a true nested dict structure
        nested = to_nested_dict(common_sheet_converted, keysep='/')
        # update the varnames of child devices
        SWDefinitionSheetFormatter.update_varnames(nested)
        return nested


class SWDefinitionSheetFormatter:
    """
    Utility class to format and process sheets from SWDefinitionTable.
    It provides methods for formatting keys, modules, and device names.
    """

    @classmethod
    def remove_excluded_sheets(
        cls, sheets: dict[str | int, pd.DataFrame]
    ) -> dict[str, pd.DataFrame]:
        """
        Removes sheets that are listed in `SWDefinitionTable`'s
        `EXCLUDED_SHEETS` from the input dictionary of sheets.

        Args:
            sheets (dict[str | int, pd.DataFrame]): The original dictionary
            containing all sheets.

        Returns:
            dict[str, pd.DataFrame]: A filtered dictionary containing only the
            sheets that are not in `EXCLUDED_SHEETS`.
        """
        excluded = SWDefinitionTable.EXCLUDED_SHEETS
        excluded_str = ", ".join(f"'{sheet}'" for sheet in excluded)
        logger.debug('Removing SW Defintion table sheets "%s".', excluded_str)
        return {
            str(sheet_name): sheet_data for
            sheet_name, sheet_data in sheets.items()
            if sheet_name not in excluded
            }

    @staticmethod
    def format_module_key(key: str) -> str:
        """
        Converts CamelCase, replaces all non-alphanumeric characters with
        underscores, removes leading/trailing underscores, and reduces multiple
        underscores to a single underscore.

        Args:
            key (str): The string to be sanitized.

        Returns:
            str: The sanitized string, transformed to lowercase and with all
            non-alphanumeric characters replaced with underscores.
        """
        # replace CamelCase words with underscore-separated words
        sanitized_key = re.sub('([a-z0-9])([A-Z])', r'\1_\2', key)
        # replace any non-alphanumeric character with an underscore
        sanitized_key = re.sub(r'[^a-z0-9_]', '_', key.lower())
        # remove leading/trailing underscores
        sanitized_key = sanitized_key.strip('_')
        # replace multiple underscores with a single underscore
        sanitized_key = re.sub('_+', '_', sanitized_key)
        # add 'general' prefix to the formatted key
        sanitized_key = f'general/{sanitized_key}'
        return sanitized_key

    @staticmethod
    def format_module_name(module_name: str) -> tuple[str, str]:
        """
        Split the module name into metamodule and module parts.

        Args:
            module_name (str): The original module name.

        Returns:
            tuple[str, str]: The metamodule and module parts of the name.
        """
        if ':' not in module_name:  # No metamodule present
            if '.' in module_name:
                metamodule = ':'.join(module_name.split('.'))
            elif '/' in module_name:
                metamodule = ':'.join(module_name.split('/'))
            else:
                metamodule = module_name
            return metamodule, module_name
        parts = module_name.split(':')
        return ':'.join(parts[:-1]), parts[-1]

    @staticmethod
    def format_module_separator(module: str) -> str:
        """
        Replace '.' with '/' in module names.

        Args:
            module (str): Original module name.

        Returns:
            str: Module name with '/' instead of '.'.
        """
        formatted = module.replace('.', '/')
        logger.debug('Formatting module name: "%s" -> "%s"', module, formatted)
        return formatted

    @classmethod
    def _format_device_names(cls, sheet: pd.DataFrame) -> pd.DataFrame:
        """
        Helper method for the `process_sheet` method to format names and
        connectors in the device data columns.

        Args:
            sheet (pd.DataFrame): The sheet DataFrame being processed.

        Returns:
            pd.DataFrame: The updated sheet DataFrame with
            formatted names and connectors.
        """
        name_key = 'general/name'
        conn_key = 'control/connector/'
        name_mask = sheet.index.get_level_values('Key') == name_key
        conn_mask = sheet.index.get_level_values('Key').str.contains(conn_key)
        column_index = 1
        while True:
            if column_index not in sheet:
                break
            name_column = sheet.loc[name_mask, column_index]
            conn_column = sheet.loc[conn_mask, column_index]
            sheet.loc[name_mask, column_index] = name_column.apply(
                cls.format_device_name)
            sheet.loc[conn_mask, column_index] = conn_column.apply(
                cls.format_device_connector)
            column_index += 1
        return sheet

    @classmethod
    def process_sheet_init(cls, sheet: pd.DataFrame) -> pd.DataFrame:
        """
        Process a DataFrame representing a sheet, dropping irrelevant data
        and formatting columns and cells.

        Args:
            sheet (pd.DataFrame): The original sheet DataFrame.

        Returns:
            pd.DataFrame: The processed sheet DataFrame.
        """
        logger.debug("Processing sheet data.")
        sheet = sheet.copy()
        # drop empty columns ...
        logger.debug("Dropping empty columns.")
        # ... where all cells are NaN
        sheet.dropna(axis='columns', how='all', inplace=True)
        # ... and where all cells are empty string values
        empty_cols = (sheet == '').all()
        sheet = sheet.drop(columns=empty_cols[empty_cols].index)
        # drop rows where either 'Key' or 'Parameter' values are empty
        logger.debug(
            "Dropping rows where both 'Key' and 'Parameter' values are empty.")
        sheet.dropna(subset=['Key', 'Parameter'], how='all', inplace=True)
        mask = (sheet['Key'] == '') & (sheet['Parameter'] == '')
        sheet = sheet[~mask]
        # format column name to int (if applicable) to maintain consistency
        logger.debug("Converting column names to integer type if applicable.")
        sheet.rename(
            columns=lambda col: int(col) if str(col).isnumeric() else col,
            inplace=True
            )
        # fill in all empty cells' values with an emplty string
        logger.debug("Filling empty cells with empty string value ('').")
        sheet = sheet.fillna('')
        # format sheet values to string to maintain consistency
        sheet = sheet.applymap(lambda cell: str(cell).strip())
        # change 'Module' values separator to '/', creare 'Metamodule' column
        logger.debug("Formatting 'Module' column values.")
        sheet[['Metamodule', 'Module']] = sheet['Module'].apply(
            lambda x: cls.format_module_name(x)).to_list()
        sheet['Module'] = sheet['Module'].apply(cls.format_module_separator)
        # split multikeys (values is 'Key' with a comma) to multiple rows
        sep = r'\s*,\s*'
        sheet = sheet.assign(Key=sheet['Key'].str.split(sep)).explode('Key')
        # substitute missing 'Key' values from the 'Parameter' values
        logger.debug(
            "Substituting missing 'Key' values from the 'Parameter' column.")
        subs_keys = sheet['Parameter'].apply(cls.format_module_key)
        sheet['Key'] = sheet['Key'].fillna(subs_keys)
        mask = sheet['Key'] == ''
        sheet.loc[mask, 'Key'] = subs_keys[mask]
        # set multiindex to 'Module' + 'Key' columns
        logger.debug("Setting multiindex to ['Module', 'Key'].")
        sheet = sheet.set_index(['Module', 'Key'])
        sheet = sheet.sort_index()
        return sheet

    @classmethod
    def process_sheet_dict_conv(cls, sheet: pd.DataFrame) -> pd.DataFrame:
        """
        Process a DataFrame representing a sheet, dropping irrelevant data
        and formatting columns and cells.

        Args:
            sheet (pd.DataFrame): The original sheet DataFrame.

        Returns:
            pd.DataFrame: The processed sheet DataFrame.
        """
        sheet = sheet.copy()
        # format "name" and "connector" the values in the device columns
        # to exclude non-alphanumerical values
        logger.debug("Formatting device names and connectors.")
        sheet = cls._format_device_names(sheet)
        # drop columns with irrelevant data
        logger.debug(
            "Dropping 'Parameter', 'Role', and 'Description' columns.")
        sheet.drop(columns=['Parameter', 'Role', 'Description'], inplace=True)
        logger.debug("Removing rows with 'notes' in the 'Key' column.")
        mask = sheet.index.get_level_values('Key').str.contains('notes')
        sheet = sheet[~mask]
        # replace 'yes' and 'no' with corresponding boolean values
        logger.debug("Converting 'yes' and 'no' values to boolean.")
        sheet = sheet.replace(['YES', 'Yes', 'yes'], True)
        sheet = sheet.replace(['NO', 'No', 'no'], False)
        return sheet

    @staticmethod
    def format_device_name(name: str, log: bool = True) -> str:
        """
        Process a string by removing all non-alphanumeric characters from
        the entire string (whitespaces and hyphens are replaced with
        underscores).

        Example:
        >>> process_string("1TC-21 *Var1")
        "1TC_21_Var1"

        Args:
            name (str): The input string to process.

        Returns:
            str: The processed string.
        """
        # Convert non-ASCII letters to their ASCII variants
        formatted = unidecode(name)
        # Replace all non-alphanumeric characters with underscores
        formatted = re.sub(r'[^a-zA-Z0-9]', '_', formatted)
        # Replace consecutive underscores with a single one
        formatted = re.sub(r'__+', '_', formatted)
        if formatted != name and log:
            logger.debug('Updated device name: "%s" -> "%s"', name, formatted)
        return formatted

    @classmethod
    def format_device_varname(cls, varname: str, log: bool = True) -> str:
        """
        Process a string by removing its leading non-alphabetical characters
        and removing all non-alphanumeric characters from the entire string
        (whitespaces and hypens are replaced with underscores).

        Example:
        >>> process_string("1TC-21 *Var1")
        "TC_21_Var1"

        Args:
            varname (str): The input string to process.

        Returns:
            str: The processed string.
        """
        # Remove all non-alphabetical leading characters
        formatted = re.sub(r'^[^a-zA-Z]+', '', varname)
        # Format the name
        formatted = cls.format_device_name(formatted, log=False)
        if formatted != varname and log:
            logger.debug(
                'Updated device varname: "%s" -> "%s"', varname, formatted
                )
        return formatted

    @classmethod
    def _update_varnames(
        cls, device_data: dict[str, dict], parent_varname: str = ''
    ) -> None:
        varname_old = device_data.get('general', {}).get('varname')
        if varname_old is None:
            # continue
            return
        if parent_varname:
            varname_new = f'{parent_varname}_{varname_old}'
            device_data['general']['varname'] = varname_new
            logger.debug(
                'Updated device varname: "%s" -> "%s"',
                varname_old, varname_new
                )
        else:
            varname_new = varname_old
        children = device_data.get('children')
        if children:
            for child_module in children.values():
                for child_device in child_module.values():
                    cls._update_varnames(child_device, varname_new)

    @classmethod
    def update_varnames(cls, nested_structure: dict[str, dict]) -> None:
        """
        Update the variable names in the nested structure by appending
        their parent varnames.

        Args:
            nested_structure (dict[str, dict]): The original nested dictionary
            structure.
        """
        special_varnames = {
            "Zone": "Zone",
            "System": "System",
            "SystemSafety": "SystemSafety"
            }
        logger.debug("Updating varnames in the nested structure.")
        for zone_name, zone_data in nested_structure.items():
            for module_name, module_data in zone_data.items():
                for device_data in module_data.values():
                    if module_name in special_varnames:
                        varname = special_varnames[module_name]
                        device_data['general']['varname'] = varname
                    else:
                        cls._update_varnames(device_data)
        logger.debug("Finished updating varnames in the nested structure.")

    @staticmethod
    def format_device_connector(connector: str, log: bool = True) -> str:
        """
        Process a string interval by removing all non-alphanumeric characters
        from each interval item (whitespaces and hyphens are replaced with
        underscores).

        Example:
        >>> format_device_connector("1TC-21 *Var1, 1TC-21 *Var2")
        "1TC_21_Var1, 1TC_21_Var1"

        Args:
            connector (str): The input string interval to process.

        Returns:
            str: The processed string interval.
        """
        try:
            interval = Interval(connector)
        except TypeError as err:
            logger.warning(
                'Could not split a connector to the parent device names "%s" '
                '(%s). The connector will be treated as a single device.',
                connector, str(err)
                )
            return SWDefinitionSheetFormatter.format_device_name(connector)
        formatted = ", ".join(
            SWDefinitionSheetFormatter.format_device_name(i) for i in interval
            )
        if formatted != connector and log:
            logger.debug(
                'Updated device connectors: "%s" -> "%s"', connector, formatted
                )
        return formatted


class SWDefinitionModule:
    """
    `SWDefinitionModule` provides utility methods for extracting data regarding
    a SW Defintion table module based on various conditions and rules (for
    instance, parent module names and connectors).

    Attributes:
        name_last (str): The last part of the full module name.
        name_full (str): The full hierarchical name of the module.
        parent_name_last (str): The last part of the full parent module name.
        parent_name_full (str): The full hierarchical name of the parent
        module.
        parent_connector (str): The connector associated with the parent
        module.
    """

    def __init__(self, module_name: str) -> None:
        """
        Initialize the SWDefinitionModule object.

        Args:
            module_name (str): The full name of the module, e.g. Signal/Symbol.
        """
        self.name_full = module_name
        self.name_last = module_name.split('/').pop()
        self.parent_name_full, self.parent_name_last = (
            self._get_parent_module_name(module_name)
            )
        self.parent_connector = (
            self.get_connector_key_for_module(self.parent_name_last)
            )

    @staticmethod
    def _get_parent_module_name(module_name: str) -> tuple[str, str]:
        """
        Internal default method to extract parent module names.

        Args:
            module_name (str): The name of the module.

        Returns:
            tuple[str, str]: The full and the last name of the parent module.
        """
        if '/' not in module_name:  # filter out devices with no parent
            return '', ''
        module_split = module_name.split('/')
        parent_module_name_full = "/".join(module_split[:-1])
        parent_module_name_last = module_split[-2]
        return parent_module_name_full, parent_module_name_last

    @classmethod
    def get_connector_key_for_module(cls, connected_module_name: str) -> str:
        """
        Generate the connector key string for the connected module. For
        example, for `connected_module_name="Gate"`, the return value is
        `'control/connector/gate'`.

        Args:
            connected_module_name (str): The name of the connected module.

        Returns:
            str: The connector key string.
        """
        if not connected_module_name:
            return ''
        if 'controller' in connected_module_name.lower():
            return 'control/connector/controller'
        return f'control/connector/{connected_module_name.lower()}'


class SWDefinitionDevice:
    """
    `SWDefinitionDevice` provides utility methods for extracting and
    manipulating data from a SW Definition device.

    Attributes:
        data (dict[str, typing.Any | dict]): The dictionary containing all
        relevant device data.
    """
    def __init__(self, device_data: dict[str, typing.Any | dict]) -> None:
        """
        Initialize the SWDefinitionDevice object with provided device data.

        Args:
            device_data (dict[str, typing.Any | dict]): The data dictionary
            containing device information.
        """
        self.data = device_data

    def get_device_name(self) -> str:
        """
        Get the device name based on available keys.

        Returns:
            str: The name of the device.

        Raises:
            KeyError: If no suitable key is found in the data dictionary.
        """
        try:
            name = self.data['general/name']
            assert isinstance(name, str)
            return name
        except KeyError as err:
            err.add_note('Device name was not found.')
            raise err


class SWDefinitionSheetToNestedDictConverter:
    """
    `SWDefinitionSheetToNestedDictConverter` is responsible for converting
    a given software definition sheet, which is in the `DataFrame` format,
    into a nested dictionary structure.
    """
    def __init__(self, sheet_data: pd.DataFrame) -> None:
        """
        Initialize the SWDefinitionSheetToNestedDictConverter object with
        the sheet data.

        Args:
            sheet_data (pd.DataFrame): The software definition sheet data.
        """
        self.data = sheet_data

    def preprocess_sheet_data(self) -> None:
        """
        Preprocess the sheet data to make it suitable for conversion into
        a nested dictionary.

        This method makes a copy of the original DataFrame stored in
        the `data` attribute, processes it using the
        the `SWDefinitionSheetFormatter.process_sheet_dict_conv` method,
        and updates the `data` attribute with the processed data.
        This preprocessing step is necessary before initializing devices and
        constructing the nested dictionary.
        """
        data = SWDefinitionSheetFormatter.process_sheet_dict_conv(self.data)
        self.data = data

    def init_devices_in_column(
        self, column_data: pd.Series
    ) -> dict[str, SWDefinitionDevice]:
        """
        Initialize devices in a given column of a DataFrame.

        This method processes a column from the whole sheet DataFrame
        to create a dictionary of SWDefinitionDevice objects. Note that
        the column data is indexed by a multiindex of the module name and
        the key as a result of preprocessing with `preprocess_sheet_data`.

        It is assumed that one column of the sheet DataFrame contains zero or
        one device data, so the created dictionary is one-to-one mapping of
        the module name keys to SWDefinitionDevice objects.

        Args:
            column_data (pd.Series): Data of a single column in the DataFrame.

        Returns:
            dict[str, SWDefinitionDevice]: A dictionary where the key is
            the module name and the value is an instance of SWDefinitionDevice.
        """
        nested_structure: dict[str, typing.Any | dict] = {}
        for (module, key), value in column_data.to_dict().items():
            if module not in nested_structure:
                nested_structure[module] = {}
            nested_structure[module][key] = value
            metatype_key = 'general/metatype'
            metatype_value = self.data['Metamodule'][module][key]
            nested_structure[module][metatype_key] = metatype_value
        return {
            module_name: SWDefinitionDevice(device_data)
            for module_name, device_data in nested_structure.items()
            }

    @staticmethod
    def update_device_name(device_data: SWDefinitionDevice) -> None:
        """
        Update the device name and variable name in a SWDefinitionDevice
        object.

        The method takes a SWDefinitionDevice object, formats its name and
        varname according to the predefined rules, and updates the object's
        data.

        Args:
            device_data (SWDefinitionDevice): The device object to be updated.
        """
        formatted_name = SWDefinitionSheetFormatter.format_device_name(
            name=device_data.get_device_name()
        )
        formatted_varname = SWDefinitionSheetFormatter.format_device_varname(
            varname=formatted_name
        )
        device_data.data['general/name'] = formatted_name
        if 'control/config/name' in device_data.data:
            device_data.data['control/config/name'] = formatted_name
        device_data.data['general/varname'] = formatted_varname

    def init_nested_structure(
        self
    ) -> dict[str, list[dict[str, typing.Any | dict]]]:
        """
        Initialize a nested structure to store device data as.

        This method iterates over device columns in the DataFrame,
        initializes devices and updates their names. Finally,
        it returns a nested dictionary containing the processed device data.

        Returns:
            dict[str, list[dict[str, typing.Any | dict]]]: A nested dictionary
            containing the device data.
        """
        nested_structure: dict[str, list] = {}
        col_idx = 1
        while True:
            if col_idx not in self.data:
                break
            devices = self.init_devices_in_column(self.data[col_idx])
            # prepare a list structure to store device data dicts
            for module_name in devices.keys():
                if module_name not in nested_structure:
                    nested_structure[module_name] = []
            # update name and varname in device data dicts
            for device_data in devices.values():
                self.update_device_name(device_data)
            for module_name, device_data in devices.items():
                if device_data.get_device_name():
                    nested_structure[module_name].append(device_data.data)
            col_idx += 1
        return nested_structure

    def remove_empty_modules(
        self,
        nested_structure: dict[str, list[dict[str, typing.Any | dict]]]
    ) -> None:
        """
        Remove modules that have no associated device data from
        the nested structure.

        This method iterates over the nested structure and identifies modules
        that do not have any device data (empty lists). It removes these
        modules from the structure to ensure the final nested dictionary only
        contains modules with relevant data.

        Args:
            nested_structure (dict[str, list[dict[str, typing.Any | dict]]]):
            The nested dictionary structure containing the module and device
            data.
        """
        to_remove = []
        for module_name, device_list in nested_structure.items():
            if not device_list:
                to_remove.append(module_name)
        for module_name in to_remove:
            del nested_structure[module_name]

    def add_child_device_data_to_parent_device_data(
        self,
        child_device: SWDefinitionDevice,
        child_module: SWDefinitionModule,
        nested_structure: dict[str, list[dict[str, typing.Any | dict]]]
    ) -> bool:
        """
        Identify the parent devices of a given child device and add the child
        device's data to the data structure of each of these parent devices.
        Note that the added child device data dictionary object is the same
        object as the original child device data dictionary.

        Args:
            child_device (SWDefinitionDevice): The child device object
            whose data needs to be added.
            child_module (SWDefinitionModule): The child module object
            associated with the child device.
            nested_structure (dict[str, list[dict[str, typing.Any | dict]]]):
            The nested dictionary structure containing all the device data.
        """
        # get the str parent device list, e.g. "par1, par3-par5",
        # and verify it is a string value
        if child_module.parent_connector not in child_device.data:
            logger.warning(
                'Device "%s" of type "%s" was identified as a potential child '
                'device, but parent connector "%s" was not found within '
                'the device data.', child_device.get_device_name(),
                child_module.name_full, child_module.parent_connector
                )
            return False
        assert isinstance(parent_device_names_str := (
            child_device.data[child_module.parent_connector]
        ), str)
        try:
            parent_device_names: typing.Iterable
            parent_device_names = Interval(parent_device_names_str)
        except TypeError:
            parent_device_names = [parent_device_names_str]
        # if a child device does not have any parent device assigned to it,
        # notify the user with a warning
        if not parent_device_names:
            child_device_name = child_device.get_device_name()
            logger.warning(
                'Device "%s" of type "%s" is not connected to any parent '
                'device of type "%s"', child_device_name,
                child_module.name_full, child_module.parent_name_full
                )
        # get the list of parent devices of the given child device
        parent_module_device_list = [
            device
            for device in nested_structure[child_module.parent_name_full]
            if device['general/name'] in parent_device_names
            ]
        # add the child device data to each parent device data
        for parent_device_data in parent_module_device_list:
            children_module_data = parent_device_data.setdefault(
                'children', {})
            children_device_data = children_module_data.setdefault(
                child_module.name_last, {})
            assert isinstance(children_device_data, dict)
            child_device_name = child_device.get_device_name()
            children_device_data[child_device_name] = child_device.data
        return True

    @staticmethod
    def copy_child_device_data(
        child_device_data: dict[str, dict | typing.Any],
        parent_module: SWDefinitionModule,
        parent_device: SWDefinitionDevice
    ) -> dict[str, dict | typing.Any]:
        parent_connector_key = SWDefinitionModule.get_connector_key_for_module(
                connected_module_name=parent_module.name_last)
        parent_device_name = parent_device.get_device_name()
        child_device_data_copy = dict(child_device_data)
        child_device_data_copy[parent_connector_key] = parent_device_name
        return child_device_data_copy

    def copy_children_device_data_within_parent_device_data(
        self,
        parent_module: SWDefinitionModule,
        parent_device: SWDefinitionDevice
    ) -> None:
        children_module_data = parent_device.data.setdefault('children', {})
        for module_data in children_module_data.values():
            assert isinstance(module_data, dict)
            for device_name, device_data in module_data.items():
                module_data[device_name] = self.copy_child_device_data(
                        device_data, parent_module, parent_device
                        )

    def format_nested_structure(
        self,
        nested_structure: dict[str, list[dict[str, typing.Any | dict]]]
    ) -> dict[str, dict]:
        """
        Integrate child device data into their respective parent devices and
        reorganize data from a list of device data to a nested dictionary with
        hierarchical device relations.

        Args:
            nested_structure (dict[str, list[dict[str, typing.Any | dict]]]):
            The initial structure with the raw device data.

        Returns:
            dict[str, dict]: The formatted nested dictionary structure.
        """
        # add child device references to their respective parent devices
        # (these child devices are removed later once they are copied)
        modules_to_remove = set()
        for module_name, device_list in nested_structure.items():
            module = SWDefinitionModule(module_name)
            if not module.parent_connector:  # is not a child module
                continue
            for device_data in device_list:
                device = SWDefinitionDevice(device_data)
                if self.add_child_device_data_to_parent_device_data(
                    device, module, nested_structure
                ):
                    modules_to_remove.add(module_name)
        # make child device data dicts under the 'children' key unique
        # (copy child device dicts)
        for module_name, device_list in nested_structure.items():
            module = SWDefinitionModule(module_name)
            for device_data in device_list:
                device = SWDefinitionDevice(device_data)
                self.copy_children_device_data_within_parent_device_data(
                    module, device
                    )
        # reorganize data to a homogenious nested dictionary from
        # a list of dictionaries
        nested_structure_formatted = {}
        for module_name, device_list in nested_structure.items():
            device_dict = {}
            for device_data in device_list:
                device_dict[device_data['general/name']] = device_data
            nested_structure_formatted[module_name] = device_dict
        # remove the original child device data
        for module_name in modules_to_remove:
            del nested_structure_formatted[module_name]
        return nested_structure_formatted

    @staticmethod
    def filter_devices(
        device_dict: dict[str, dict],
        preserve_devices: typing.Iterable[str],
        raise_if_filtered_all: bool = False
    ) -> dict[str, dict]:
        """
        Filter the devices in the dictionary based on a list of device names
        to preserve.

        Args:
            device_dict (dict[str, dict]): The dictionary containing device
            data.
            preserve_devices (typing.Iterable[str]): An iterable of device
            names to be preserved.
            raise_if_filtered_all (bool): A flag to raise an error if
            all devices are filtered out.

        Returns:
            dict[str, dict]: A dictionary containing only the data of devices
            listed in preserve_devices.
        """
        filtered_data = {}
        for device_name, device_data in device_dict.items():
            if device_name in preserve_devices:
                filtered_data[device_name] = device_data
        # if there were devices to preserve, but none of the devices in
        # the provided device_dict matched, and the filetered_data is empty
        if not filtered_data and raise_if_filtered_all and preserve_devices:
            devices = ", ".join(preserve_devices)
            raise KeyError(f'None of the devices "{devices}" were found')
        return filtered_data

    def to_dict(self) -> dict[str, dict]:
        """
        Convert the sheet data into a nested dictionary structure.

        This method acts as the main interface for converting the sheet data
        into a nested dictionary. It sequentially calls other methods
        to preprocess the data, initialize the nested structure, remove empty
        modules, and format the structure, resulting in a nested dictionary
        representation of the sheet data.

        Returns:
            dict[str, dict]: The nested dictionary structure containing
            the processed device data.
        """
        self.preprocess_sheet_data()
        devices_list = self.init_nested_structure()
        self.remove_empty_modules(devices_list)
        devices_dict = self.format_nested_structure(devices_list)
        return devices_dict


class SWDefinitionCommonSheetToNestedDictConverter(
    SWDefinitionSheetToNestedDictConverter
):
    def to_dict(self) -> dict[str, dict]:
        """
        Raises a `NotImplementedError` to indicate that the `to_dict_common`
        method should be used for processing "Common" sheets instead of
        the regular `to_dict` method from the parent class. Such sheets have
        specific processing requirements that  are handled by
        the `to_dict_common` method  with a different signature.

        Raises:
            NotImplementedError: Indicates the use of `to_dict_common`
            for "Common" sheets.
        """
        raise NotImplementedError(
            '"to_dict_common" method must be used to process the Common sheet'
            )

    def update_system_data(
        self, system_name: str, common_data: dict
    ) -> None:
        """
        Update the system data within the "Common" sheet data structure.

        This method updates various attributes of the system data, such as
        'varname', 'control/function', 'control/connector', and
        'control/config/name', based on the provided system name.

        Args:
            system_name (str): The name of the system to be updated.
            common_data (dict): The "Common" sheet data structure where
            the system data is located.
        """
        system_data = common_data['System'][system_name]
        system_data['general/varname'] = 'System'
        system_data['control/function'] = 'fSystem_System'
        system_data['control/devType'] = 'SystemType'
        system_data['control/connector'] = {}
        system_data['control/config/name'] = system_name
        system_safety_data = system_data.get('Safety')
        if system_safety_data:
            system_safety_device_data = system_safety_data['SystemSafety']
            system_safety_device_data['control/function'] = 'fSafety_SafetyCPU'

    def get_project_name(self, common_data: dict) -> str:
        """
        Extract the project name from the common data structure.

        Args:
            common_data (dict): The common data structure containing
            the project information.

        Returns:
            str: The extracted project name.
        """
        project_data: dict = common_data['Project']
        project_key: str = list(project_data.keys()).pop()
        project_name = project_key.split('/').pop()
        return project_name

    def update_shv_device_id(
        self, project_name: str, system_name: str, common_data: dict
    ) -> None:
        """
        Update the SHV device ID in the common data structure.

        This method constructs and updates the SHV "deviceID" key using
        the project name and system name.

        Args:
            project_name (str): The name of the project.
            system_name (str): The name of the system.
            common_data (dict): The common data structure where
            the SHV device ID is to be updated.
        """
        broker = f'{project_name}{system_name}-broker'
        try:
            common_data['SHV']['SHV']['control/config/deviceID'] = broker.lower()
        except KeyError:
            logger.warning('Failed to update SHV Device ID broker name.')

    def to_dict_common(
        self, system_name: str, commfunc_names: list[str]
    ) -> dict[str, dict]:
        """
        Convert the "Common" sheet data into a nested dictionary structure for
        a specific system.

        This method extends the `to_dict` method from the parent class to
        include additional processing steps specific to "Common" sheets.
        It filters and updates system and communication function data based on
        the provided system name and communication function names.

        Args:
            system_name (str): The name of the system to be processed.
            commfunc_names (list[str]): A list of communication function names
            used within the system.

        Returns:
            dict[str, dict]: The nested dictionary structure containing the
            processed "Common" sheet data.
        """
        data = super().to_dict()
        system_name_formatted = (
            SWDefinitionSheetFormatter.format_device_name(system_name)
            )
        if 'System' in data:
            data['System'] = self.filter_devices(
                data['System'],
                [system_name_formatted],
                raise_if_filtered_all=True
                )
        if 'CommFunction' in data:
            data['CommFunction'] = self.filter_devices(
                data['CommFunction'],
                commfunc_names
            )
        self.update_system_data(system_name_formatted, data)
        project_name = self.get_project_name(data)
        self.update_shv_device_id(project_name, system_name_formatted, data)
        return data



from g3hardware import format_type, is_isle_head_type
from g3tables import PLCCompositionIOTable

from ._logger import logger
from .type_hinting import HWIsleDict, HWModuleDict


def get_hardware_connections(table: PLCCompositionIOTable) -> list[HWIsleDict]:
    connections: list[HWIsleDict] = []
    isle: HWIsleDict | None = None
    logger.info('Extracting hardware connections from PLCCompositionIOTable')
    for module_type, cabinet, name_suffix in table.iter_plc_units():
        module_type = format_type(module_type)
        module_dict: HWModuleDict = {
            'type': module_type,
            'cabinet': cabinet,
            'name_suffix': name_suffix
            }
        if is_isle_head_type(module_type):
            isle = {'head': module_dict, 'tail': []}
            connections.append(isle)
        elif isle is None:
            continue
        else:
            isle['tail'].append(module_dict)
    return connections



import argparse
import functools
import os
import logging
import typing

from g3core import G3Core

from ..plc_composition_io_table import PLCCompositionIOTable
from ..sw_definition_table import SWDefinitionTable
from ..visualization_table import VisualizationTable
from ..utils.gdrive import list_tables_from_gdrive
from ._system_config_processor import SystemConfigProcessor
from ._logger import logger


class InputArgValidators:

    @classmethod
    def gdrive_folder(cls, folder: typing.Any) -> tuple[str, str | None]:
        """folder@system -> folder, system"""
        folder = cls.is_str(folder, 'Google Drive folder')
        folder_data: list[str] = folder.split('@')
        if len(folder_data) == 1:
            folder, system = folder_data[0].strip(), None
        elif len(folder_data) == 2:
            folder, system = folder_data[0].strip(), folder_data[1].strip()
        else:
            raise ValueError('Invalid Google Drive folder format.')
        if not folder:
            raise ValueError('Google Drive folder name cannot be empty.')
        if not system:
            raise ValueError('System name cannot be empty.')
        return folder, system

    @classmethod
    def sw_table_path(cls, path: typing.Any) -> tuple[str, str]:
        """path@system -> path, system"""
        path = cls.is_str(path, 'SW Definition table path')
        path_data: list[str] = path.split('@')
        if len(path_data) != 2:
            raise ValueError('Invalid SW Definition table path format.')
        path, system_name = path_data[0].strip(), path_data[1].strip()
        if not path:
            raise ValueError('SW Definition table path cannot be empty.')
        if not system_name:
            raise ValueError('System name cannot be empty.')
        return path, system_name

    @classmethod
    def io_table_path(cls, path: typing.Any) -> tuple[str, set[str]]:
        """path@cabinet1+cabinet2 -> path, [cabinet1, cabinet2]"""
        path = cls.is_str(path, 'PLC Composition IO table path')
        path_data: list[str] = path.split('@')
        if len(path_data) == 1:
            path = path_data[0].strip()
            sheets = set()
        elif len(path_data) == 2:
            path = path_data[0].strip()
            sheets = set((sh.strip() for sh in path_data[1].split('+') if sh))
        else:
            raise ValueError('Invalid PLC Composition IO table path format.')
        if not path:
            raise ValueError('PLC Composition IO table path cannot be empty.')
        return path, sheets

    @classmethod
    def visu_table_path(cls, path: typing.Any) -> str:
        """path -> path"""
        path = cls.is_str(path, 'Visualization table path')
        if not path:
            raise ValueError('Visualization table path cannot be empty.')
        return path

    @classmethod
    def g3core_path(cls, path: typing.Any) -> tuple[str, str]:
        if path is None:
            return 'remote', 'master'  # default fallback
        path = cls.is_str(path, 'G3Core library path')
        path_data: list[str] = path.split('@')
        if len(path_data) == 1:
            path = path_data[0].strip()
            if path == 'remote':
                return path, 'master'
            elif os.path.isdir(path):
                return path, ''
            raise ValueError(
                f'Specified local G3Core library directory "{path}" '
                f'does not exist.'
                )
        elif len(path_data) == 2:
            path, branch = path_data[0].strip(), path_data[1].strip()
            if path != 'remote':
                raise ValueError(
                    'Local G3Core directory path got an unexpected branch '
                    'specification. An @-separated path can only be used with '
                    'the "remote" prefix, e.g., "remote@master".'
                    )
            return path, branch
        else:
            raise ValueError('Invalid G3Core directory path format.')

    @staticmethod
    def is_str(arg: typing.Any, arg_name: str) -> str:
        if not isinstance(arg, str):
            raise TypeError(
                f'Invalid {arg_name} type: "{type(arg).__name__}" '
                f'(expected type "str").'
                )
        return arg

    @classmethod
    def is_system_name_valid(
        cls, name_sw_table: str, name_gdrive: str
    ) -> str | None:
        if name_sw_table and name_gdrive:
            if name_sw_table == name_gdrive:
                return cls.is_str(name_sw_table, 'system name')
            raise ValueError(
                f'Specified system name is ambiguous (SW Definition table: '
                f'"{name_sw_table}", Google Drive folder: "{name_gdrive}").'
                )
        if not name_sw_table and not name_gdrive:
            return None
        if name_sw_table and not name_gdrive:
            return cls.is_str(name_sw_table, 'system name')
        if not name_sw_table and name_gdrive:
            return cls.is_str(name_gdrive, 'system name')
        raise ValueError(
            f'Unexpected system name (SW Definition table: '
            f'"{name_sw_table}", Google Drive folder: "{name_gdrive}").'
            )  # should not happen


class InputArgs(typing.TypedDict):
    gdrive_folder: str | None
    sw_table_path: str | None
    system_name: str | None
    io_table_path: str | None
    io_table_sheets: set[str] | None
    visu_table_path: str | None
    g3core_path: str
    g3core_branch: str
    update: bool
    output_path: str
    log_level: str


def get_input_args() -> InputArgs:
    parser = argparse.ArgumentParser(
        description=(
            'Create a JSON representation of a G3 System configuration data.'
            )
        )
    parser.add_argument(
        '-g',
        '--gdrive_folder',
        type=InputArgValidators.gdrive_folder,
        required=False,
        default=(None, None),
        help=(
            'Remote Google Drive repository folder name (e.g., "Prg015") '
            'with Project G3 project tables. The tables in the folder are '
            'identified automatically. Any automatically collected table name '
            'is overridden by a corresponding explicitly provided argumement. '
            'Format as follows: "<gdrive_folder>@<system_name>", e.g., '
            '"Prg015@Zone01".'
            )
        )
    parser.add_argument(
        '-sw',
        '--swtable-path',
        type=InputArgValidators.sw_table_path,
        required=False,
        default=(None, None),
        help=(
            'Path to the SW Definition Table. The path ending with ".xlsx" '
            'is considered to be a local file path. Otherwise, the path '
            'is considered to be a Google Drive table name. If the name '
            'contains whitespace, it should be enclosed in quotes. '
            'Format as follows: "<table_path>@<system_name>"., e.g., '
            '"SWDef_Prg015.xlsx@Zone01" or '
            '"Prg015 G3 project SW definition EID00012345"@Zone01.'
            )
        )
    parser.add_argument(
        '-io',
        '--iotable-path',
        type=InputArgValidators.io_table_path,
        required=False,
        default=(None, []),
        help=(
            'Path to the PLC Composition and IO table. The path ending with '
            '".xlsx" is considered to be a local file path. Otherwise, '
            'the path is considered to be a Google Drive table name. If '
            'the name contains whitespace, it should be enclosed in quotes.'
            'If the cabinet names are not provided, they are extracted from '
            'the SW Definition Table data. '
            'Format as follows: '
            '1) "<table_path>" (no sheet names provided, e.g., '
            '"Prg015_HW.xlsx" or '
            '"Prg015 PLC composition and IO Table EID00012345"); '
            '2) "<table_path>@<cabinet_name>" (one sheet name provided, e.g., '
            '"Prg015_HW.xlsx@SIG01" or '
            '"Prg015 PLC composition and IO Table EID00012345"@SIG01); '
            '3) "<table_path>@<cabinet_name1>+<cabinet_name2>+<cabinet_name3>"'
            ' (two or more sheet names provided, e.g., ).'
            '"Prg015_HW.xlsx@SIG01+SIG02" or '
            '"Prg015 PLC composition and IO Table EID00012345"@SIG01+SIG02); '
            )
        )
    parser.add_argument(
        '-v',
        '--visutable-path',
        type=InputArgValidators.visu_table_path,
        required=False,
        default=None,
        help=(
            'Path to the Visualization table. The path ending with ".xlsx" is '
            'considered to be a local file path. Otherwise, the path is '
            'considered to be a Google Drive table name. If the name contains '
            'whitespace, it should be enclosed in quotes. Format as follows: '
            '"<table_path>", e.g., "Visu_Prg015.xlsx" or '
            '"Prg015 G3 project Visualization EID00012345".'
            )
        )
    parser.add_argument(
        '-c',
        '--g3core-path',
        type=InputArgValidators.g3core_path,
        required=False,
        default=('remote', 'master'),
        help=(
            'Path to the G3 Core library files. Format as follows: '
            '1) "<local_path>", e.g., "C:/.../SystemG3Core"; '
            '2) "remote" (default brach "master" is used); '
            '3) "remote@<branch>" (the branch may be referred by its name or '
            'by its hash), e.g., "remote@master", "remote@Mil014", or '
            '"remote@f02ddf805d25c2f8d61edd11a06f8e28e4b447bd".'
            )
        )
    parser.add_argument(
        '-u',
        '--update',
        action='store_true',
        help=(
            'Update an existing JSON configuration file, only adding new '
            'key-value pairs and preserving existing data.'
            )
        )
    parser.add_argument(
        '-o',
        '--output-path',
        type=str,
        required=False,
        default=os.getcwd(),
        help='Path to the directory to save the JSON configuration file to.'
        )
    parser.add_argument(
        '--log-level',
        type=str,
        required=False,
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help='Set the logging level.'
        )
    args = parser.parse_args()
    sw_table_path, sw_table_system_name = args.swtable_path
    io_table_path, io_table_sheets = args.iotable_path
    g3core_path, g3core_branch = args.g3core_path
    gdrive_folder, gdrive_system_name = args.gdrive_folder
    try:
        system_name = InputArgValidators.is_system_name_valid(
            sw_table_system_name, gdrive_system_name
            )
    except Exception as e:
        logger.error(e)
        raise SystemExit from e
    return {
        'gdrive_folder': gdrive_folder,
        'sw_table_path': sw_table_path,
        'system_name': system_name,
        'io_table_path': io_table_path,
        'io_table_sheets': io_table_sheets,
        'visu_table_path': args.visutable_path,
        'g3core_path': g3core_path,
        'g3core_branch': g3core_branch,
        'update': args.update,
        'output_path': args.output_path,
        'log_level': args.log_level
        }


class InputArgsProcessor:
    def __init__(self, args: InputArgs) -> None:
        self._args = args

    @functools.cached_property
    def g3core(self) -> G3Core:
        g3core_path = self._args['g3core_path']
        g3core_branch = self._args['g3core_branch']
        if g3core_path == 'remote':
            return G3Core.from_giltab(branch=(g3core_branch or 'master'))
        if g3core_branch:
            raise ValueError(
                'Local G3Core directory path got an unexpected branch '
                'specification. An @-separated path should be used '
                'with the "remote" prefix, e.g., "remote@master".'
                )
        return G3Core.from_local(g3core_path)

    @functools.cached_property
    def gdrive_tables(self) -> list[str]:
        gdrive_folder = self._args['gdrive_folder']
        if not gdrive_folder:
            return []
        return list_tables_from_gdrive(project=gdrive_folder)

    @functools.cached_property
    def sw_table(self) -> SWDefinitionTable | None:
        table_path = self._args['sw_table_path']
        if table_path:
            return SWDefinitionTable.load(table_path)
        for table in self.gdrive_tables:
            if SWDefinitionTable.is_name_valid(table):
                return SWDefinitionTable.from_gdrive(table)
        return None

    @functools.cached_property
    def io_sheets(self) -> set[str]:
        # check if io sheets are provided by the user
        if self._args['io_table_sheets']:
            return self._args['io_table_sheets']
        # try parsing the sheets (as cabinet names) from the sw table
        system = self._args['system_name']
        sheets: set[str] = set()
        if not self.sw_table or not system:
            return sheets
        zone_names = self.sw_table.get_system_sheets(system).keys()
        for zone_name in zone_names:
            if zone_name == 'Common':
                continue
            for cabinet_name in self.sw_table.get_cabinet_names(zone_name):
                sheets.add(cabinet_name)
        return sheets

    @functools.cached_property
    def io_table(self) -> PLCCompositionIOTable | None:
        if not self.io_sheets:
            return None
        table_path = self._args['io_table_path']
        if table_path:
            return PLCCompositionIOTable.load(table_path, self.io_sheets)
        for table in self.gdrive_tables:
            if PLCCompositionIOTable.is_name_valid(table):
                return PLCCompositionIOTable.from_gdrive(table, self.io_sheets)
        return None

    @functools.cached_property
    def visu_table(self) -> VisualizationTable | None:
        table_path = self._args['visu_table_path']
        if table_path:
            return VisualizationTable.load(table_path)
        for table in self.gdrive_tables:
            if VisualizationTable.is_name_valid(table):
                return VisualizationTable.from_gdrive(table)
        return None


def main() -> None:
    args = get_input_args()
    logging.basicConfig(
        level=args['log_level'],
        format='[%(name)s] %(levelname)s:%(message)s'
        )
    args_processor = InputArgsProcessor(args)
    data_processor = SystemConfigProcessor(
        system_name=args['system_name'],
        sw_table=args_processor.sw_table,
        io_table=args_processor.io_table,
        visu_table=args_processor.visu_table,
        g3core=args_processor.g3core
        )
    os.makedirs(args['output_path'], exist_ok=True)
    data_processor.to_file(
        dirpath=args['output_path'],
        filename='SystemConfig.json',
        update_if_exists=args['update']
        )



from .plc_composition_io_table._table import PLCCompositionIOTable
from .sw_definition_table._table import SWDefinitionTable
from .visualization_table._table import VisualizationTable
from . import (
    plc_composition_io_table,
    sw_definition_table,
    visualization_table,
    system_config,
    utils
)

__all__ = [
    'SWDefinitionTable',
    'PLCCompositionIOTable',
    'VisualizationTable',
    'plc_composition_io_table',
    'sw_definition_table',
    'visualization_table',
    'system_config',
    'utils'
    ]



from ._sw_system_dict_wrapper import SWSystemDictWrapper
from ._system_config_processor import SystemConfigProcessor
from . import type_hinting

__all__ = ['SWSystemDictWrapper', 'SystemConfigProcessor', 'type_hinting']
